#include <string.h>
#include "forge.h"
#include "imagine.h"
#define TESTIFY_INTERNAL
#include "testify.h"

const char *testify_discovery_global_character_transform_new = "zxvtsrqpnmkjhgdb";
const char *testify_discovery_global_character_transform = "lfcmugypwbvkjxqz";
#define TESTIFY_DISCOVER_USER_NAME_LENGTH 12

const char *testify_discovery_irc_character_transform = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_|";
const unsigned char testify_discovery_irc_character_inv_transform[] = {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 255, 255, 255, 255, 255, 255, 255, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 255, 255, 255, 255, 62, 255, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 255, 63, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255};

void testify_discover_global_encode(char *output, uint8 *input, uint input_count, boolean terminate)
{	
	uint i, j, lookup;
	for(i = j = 0; i < input_count;)
	{
		lookup = input[i] & 63;
		output[j++] = testify_discovery_irc_character_transform[lookup];
		lookup = (input[i++] >> 6) & 3;
		if(i == input_count)
		{
			output[j++] = testify_discovery_irc_character_transform[lookup];
			break;
		}
		lookup |= (input[i] << 2) & 60;
		output[j++] = testify_discovery_irc_character_transform[lookup];
		lookup = (input[i++] >> 4) & 15;
		if(i == input_count)
		{
			output[j++] = testify_discovery_irc_character_transform[lookup];
			break;
		}
		lookup |= (input[i] << 4) & 48;
		output[j++] = testify_discovery_irc_character_transform[lookup];
		lookup = (input[i++] >> 2) & 63;
		output[j++] = testify_discovery_irc_character_transform[lookup];
	}
	if(terminate)
		output[j] = 0;
}




uint testify_discover_global_decode(char *output, char *input, uint output_max)
{
	uint i, j;
	uint8 read;
	for(i = j = 0; i < output_max; i++)
	{	
		read = testify_discovery_irc_character_inv_transform[input[j++]];
		if(read == 255)
			return i;
		output[i] = read;
		read = testify_discovery_irc_character_inv_transform[input[j++]];
		if(read == 255)
			return i + 1;
		output[i++] |= (read << 6) & (128 + 64);
		if(i == output_max)
			return i;
		output[i] = (read >> 2) & 15;
		read = testify_discovery_irc_character_inv_transform[input[j++]];
		if(read == 255)
			return i + 1;
		output[i++] |= (read << 4) & (128 + 64 + 32 + 16);
		if(i == output_max)
			return i;
		output[i] = (read >> 4) & 3;
		read = testify_discovery_irc_character_inv_transform[input[j++]];
		if(read == 255)
			return i + 1;
		output[i] |= (read << 2) & (128 + 64 + 32 + 16 + 8 + 4);
	}
	return i;
}

void testify_discover_irc_print()
{
	uint i, j;
	for(i = 0; i < 256; i++)
	{
		for(j = 0; j < 64 && testify_discovery_irc_character_transform[j] != i; j++);
		if(j < 64)
			printf("%u, ", j);
		else
			printf("255, ");
	}
	{
		char *hello = "hello world";
		char output[256], decoded[256];
		testify_discover_global_encode(output, hello, 12, TRUE);
		printf("encoded: -%s-", output);
		i = testify_discover_global_decode(decoded, output, 256);
		printf("decoded: -%s-", decoded);
		i = 0;
	}
}

typedef enum{
	TESTIFY_DIC_LISTENER_ONLY,
	TESTIFY_DIC_SERVICE,
	TESTIFY_DIC_SERVER,
	TESTIFY_DIC_COUNT
}TestifyDiscoverIRCCommands;

void testify_discover_irc_name(TestifyDiscovery *discover, TestifyAddress *address, char name[(TESTIFY_ADDRESS_MAX_SERIALIZE_SIZE / 3 + 1) * 4 + 1])
{
	uint8 serialized[TESTIFY_ADDRESS_MAX_SERIALIZE_SIZE + 1];
	TestifyHandle *h;
	size_t size = 1;
	uint i;
	h = discover->service_handle;
	if(discover->global_handle != address && (discover->method_flags & TESTIFY_DMF_BROADCAST_WAN))
	{
		serialized[0] = TESTIFY_DIC_SERVER;
	}else if(discover->message[0] != 0)
	{
		serialized[0] = TESTIFY_DIC_SERVICE;
	}else
		serialized[0] = TESTIFY_DIC_LISTENER_ONLY;
	size += testify_address_serialize(address, &serialized[1]);
	testify_discover_global_encode(name, serialized, size, TRUE);
}


TestifyHandle *testify_discover_irc_connect(TestifyDiscovery *discover, char *dns, TestifyAddress *address, uint8 *service)
{
	char name[(TESTIFY_ADDRESS_MAX_SERIALIZE_SIZE / 3 + 1) * 4 + 1];
	char channel[2 + 8 * 4 / 3];
	TestifyServiceState status;
	uint i;
	testify_discover_irc_name(discover, address, name);
	testify_discover_global_encode(channel, service, 8, TRUE);
	if(channel[0] >= '0' && channel[0] <= '9')
		channel[0] += 'A' - '0';
	return testify_irc_create(dns, name, channel);
}

/*
void testify_discover_irc_send(TestifyDiscovery *discover)
{
	char buffer[9];
	uint i, j;
	memcpy(buffer, "PRIVMSG #", 9);
	testify_discover_global_encode(&buffer[9], discover->service, 8, TRUE);
	if(buffer[9] >= '0' && buffer[9] <= '9')
		buffer[9] += 'A' - '0';
	for(i = 0; buffer[i] != 0; i++);
	buffer[i++] = ' ';
	buffer[i++] = ':';
	for(j = 0; discover->message[j] != 0; j++);
	testify_discover_global_encode(&buffer[i], discover->message, j, TRUE);
	buffer[i++] = '\r';
	buffer[i++] = '\n';
	buffer[i++] = '\0';
	sprintf(buffer, "PRIVMSG #%s :%s\r\n", channel);
	testify_discover_global_encode(channel, service, 8, TRUE);
	discover->service
	discover->message
}*/

void testify_discover_irc_message_send(TestifyDiscovery *discover, char *sender)
{
	char message[(256 / 3) * 4 + 2];
	uint i;
	for(i = 0; discover->message[i] != 0; i++);
	testify_discover_global_encode(message, discover->message, i, TRUE);
	testify_irc_channel_message_send(discover->irc_handle, 0, message);
}


void testify_discover_irc_command_func(uint command, char *sender, char *channel, char *message, void *user)
{
	TeestifyDiscoverPeerInternal *peer;
	uint8 buffer[32], decoded[256];
	TestifyDiscovery *discover;
	TestifyAddress address;
	char address_print[TESTIFY_ADDRESS_MAX_STRING_SIZE], *nick;
	uint i, size;
//	printf("testify_discover_irc_command_func %u\n", (uint)command);
	if(command != TESTIFY_IRC_COMMAND_QUIT &&
		command != TESTIFY_IRC_COMMAND_PART &&
		command != TESTIFY_IRC_COMMAND_PRIVMSG &&
		command != TESTIFY_IRC_COMMAND_NOTICE &&
		command != TESTIFY_IRC_COMMAND_JOIN)
			return;

	if(sender == NULL)
		return;
	discover = user;

	nick = testify_irc_nick_get(discover->irc_handle);
	for(i = 0; sender[i] != 0 && nick[i] == sender[i]; i++);
	if(nick[i] == sender[i])
	{
	//	printf("Nick: %s %s", nick, sender);
		if(command == TESTIFY_IRC_COMMAND_JOIN)
			testify_discover_irc_message_send(discover, sender);
		return;
	}
	size = testify_discover_global_decode(buffer, sender, 32);
	if(size == 0 || buffer[0] == TESTIFY_DIC_LISTENER_ONLY)
		return;
	if(0 == testify_address_deserialize(&address, &buffer[1], size - 1))
		return;

	
	if(address.transport == TESTIFY_TRANSPORT_NULL)
	{
		testify_discover_irc_message_send(discover, sender);
		return;	
	}
	testify_address_print(&address, address_print); 
//	printf("name parsed as: %s command %u command %u\n", address_print, (uint)buffer[0], command);

	if(buffer[0] == TESTIFY_DIC_SERVER)
	{
		testify_status_service_set(TESTIFY_SST_MASTER_SERVER, "IRC derived", &address);
		return;
	}
	if(buffer[0] == TESTIFY_DIC_SERVICE)
	{
		switch(command)
		{
			case TESTIFY_IRC_COMMAND_QUIT :
			case TESTIFY_IRC_COMMAND_PART :
				peer = testify_discover_peer_lookup(discover, &address, FALSE);
				if(peer != NULL)
					testify_discover_peer_delete(discover,  peer);
			break;
			case TESTIFY_IRC_COMMAND_PRIVMSG :
			case TESTIFY_IRC_COMMAND_NOTICE : 
				size = testify_discover_global_decode(decoded, message, 256 - 1);
				decoded[size] = 0;
				peer = testify_discover_peer_update(discover, NULL, &address, decoded, TESTIFY_DMF_WAN_IRC, ~0);
			break;
			case TESTIFY_IRC_COMMAND_JOIN :
				testify_discover_irc_message_send(discover, sender);
				peer = testify_discover_peer_update(discover, NULL, &address, NULL, TESTIFY_DMF_WAN_IRC, ~0);
			break;
		}
	}
}
