#include <stdio.h>
#include "imagine.h"
#define TESTIFY_INTERNAL
#include "testify.h"

#define TESTIFY_LOCAL_TIMER 30.0

extern void testify_discover_global_update(TestifyHandle* handle, char *service, uint16 service_port, char *message, char *buffer, uint *buffer_used, uint buffer_size);

#define TESTIFY_DISCOVER_USER_ID_LENGTH 48

void testify_obfuscate(uint8 *key, uint8 *data, uint length, uint8 salt)
{
	uint8 registers[8], pos, a, b;
	uint i;
	for(i = 0; i < 8; i++)
		registers[i] = key[i] ^ salt;
	registers[0] ^= length;
	for(i = 0; i < length; i++)
	{
		pos = registers[i & 7];
		a = pos & 7;
		b = a + 1 + ((pos >> 4) & 3);
		data[i] ^= registers[a] ^ registers[b & 7];
		registers[a] ^= ((pos >> 1) & 127) | ((pos << 7) & 128);
	}
}

TeestifyDiscoverPeerInternal *testify_discover_peer_lookup(TestifyDiscovery *discovery, TestifyAddress *address, boolean local)
{
	uint i;
	if(local)
	{
		for(i = 0; i < discovery->peer_count; i++)
			if(testify_address_compare(&discovery->peers[i].address_local, address))
				return &discovery->peers[i];
	}else
		for(i = 0; i < discovery->peer_count; i++)
			if(testify_address_compare(&discovery->peers[i].address_global, address))
				return &discovery->peers[i];
	return NULL;
}

void testify_discover_peer_delete(TestifyDiscovery *discovery,  TeestifyDiscoverPeerInternal *peer)
{
	discovery->peers[peer - discovery->peers] = discovery->peers[--discovery->peer_count];
}

TeestifyDiscoverPeerInternal *testify_discover_peer_update(TestifyDiscovery *discovery, TestifyAddress *address_local, TestifyAddress *address_global, char *message, uint type, uint timeout)
{
	TeestifyDiscoverPeerInternal *peer = NULL;
	uint i, j;
	
	if(address_local != NULL)
		peer = testify_discover_peer_lookup(discovery, address_local, TRUE);
	if(peer == NULL && address_global != NULL)
		peer = testify_discover_peer_lookup(discovery, address_global, FALSE);

	if(peer != NULL)
	{
		if(address_global != NULL && peer->address_global.transport == TESTIFY_TRANSPORT_NULL)
			peer->address_global = *address_global;		
		if(address_local != NULL && peer->address_local.transport == TESTIFY_TRANSPORT_NULL)
			peer->address_local = *address_local;			
		peer->timeout = 0;		
		peer->peer.method |= type;
		if(message != NULL)
		{
			for(j = 0; message[j] != 0 && peer->peer.message[j] == message[j]; j++);
			if(peer->peer.message[j] != message[j])
			{
				for(j = 0; j < TESTIFY_NETWORK_PEER_MESSAGE_LENGTH - 1 && message[j] != 0; j++)
					peer->peer.message[j] = message[j];
				peer->peer.message[j] = 0;
				discovery->updated = TRUE;
			}
		}
		return peer;
	}
	if(discovery->peer_count == discovery->peer_allocted)
	{
		discovery->peer_allocted += 16;
		discovery->peers = realloc(discovery->peers, (sizeof *discovery->peers) * discovery->peer_allocted);
	}
	peer = &discovery->peers[discovery->peer_count++];
	if(address_local != NULL)
		peer->peer.address = peer->address_local = *address_local;
	else
		peer->address_local.transport = TESTIFY_TRANSPORT_NULL;
	if(address_global != NULL)
		peer->peer.address = peer->address_global = *address_global;
	else
		peer->address_global.transport = TESTIFY_TRANSPORT_NULL;

	if(message != NULL)
	{
		for(i = 0; message[i] != 0 && i < TESTIFY_NETWORK_PEER_MESSAGE_LENGTH - 1; i++)
			peer->peer.message[i] = message[i];
		peer->peer.message[i] = 0;
	}else
		peer->peer.message[0] = 0;
	peer->discover_port = 0;
	peer->user_id[0] = 0;
	peer->timeout = timeout;
	peer->peer.method = type;
	discovery->updated = TRUE;
	return peer;
}


void testify_discover_update(TestifyDiscovery *discover, double delta) 
{
	uint i, selected;
	TestifyDiscovery **d;
	TestifyAddress address;
	char service_name[256];
	discover->obfuscate_salt ^= 1;
	if(discover->local_handle != NULL)
		testify_discover_broadcast_update(discover, delta);
	if(discover->global_handle != NULL)
	{
		if(testify_connected(discover->global_handle))
		;//	testify_discover_global_update(discover->global_handle, discover->service, discover->service_port, discover->message, discover->global_buffer, &discover->global_buffer_used, TESTIFY_DISCOVER_GLOBAL_BUFFER_SIZE);
		else
		{
			printf("Global connection disconnected\n");
			testify_destroy(discover->global_handle);
			discover->global_handle = NULL;
			discover->global_buffer_used = 0;
		//	discover->global_handle = testify_discover_global_connect(discover->service);
		}
	}
	
	if(discover->method_flags & TESTIFY_DMF_WAN_IRC)
	{
		if(discover->irc_handle == NULL)
		{
			for(i = 0; testify_address(discover->service_handle, i, &address); i++)
			{
				if(address.transport == TESTIFY_TRANSPORT_IPV4)
				{
					TestifyAddressType address_type;
					address_type = testify_address_type(&address);
					if(TESTIFY_AT_GLOBAL == address_type)
					{
						selected = testify_status_service_select(TESTIFY_SST_IRC, service_name, 256, NULL, TESTIFY_SS_INACTIVE);
						if(selected != ~0)
						{ 
							discover->irc_handle = testify_discover_irc_connect(discover, service_name, &address, &discover->obfuscate_key);								
							if(discover->irc_handle != NULL)
							{
								testify_status_service_status_set(TESTIFY_SST_IRC, service_name, TESTIFY_SS_SUCSESS);
							}else
								testify_status_service_status_set(TESTIFY_SST_IRC, service_name, TESTIFY_SS_FAILED);
						}
						break;
					}
				}
			}
		}else
		{
			TestifyHandle	*irc_handle;
			int socket;
			irc_handle = discover->irc_handle;
			socket = discover->irc_handle->socket;
			testify_irc_parser(discover->irc_handle, testify_discover_irc_command_func, discover);
		}
	}
}

extern void testify_status_initialize();

TestifyDiscovery *testify_discover_create(char *service, TestifyHandle *handle, char *message, uint32 method_flags)
{
	TestifyDiscovery *discovery;
	TestifyAddress address;
	union{ uint8 bytes[2]; uint16 port;}port_select;
	uint8 *p;
	uint i;
	
	testify_status_initialize();
	discovery = malloc(sizeof *discovery);
	discovery->peers = NULL;
	discovery->peer_count = 0;
	discovery->peer_allocted = 0;
	discovery->service_handle = handle;
	p = discovery;
	port_select.port = 0;
	discovery->obfuscate_salt = 0;
	for(i = 0; i < (sizeof discovery); i++)
		discovery->obfuscate_salt ^= p[i];
	for(i = 0; i < 8; i++)
		discovery->obfuscate_key[i] = 0;
	for(i = 0; service[i] != 0 && i < 64 - 1; i++)
	{
		discovery->service[i] = service[i];
		port_select.bytes[i % 2] ^= service[i];
		discovery->obfuscate_key[i & 7] += service[i];
	}
	port_select.port |= 0x8000;
	if(handle != NULL)
		if(port_select.port == handle->address.port)
			port_select.port++;
	discovery->service[i] = 0;
	for(i = 0; message[i] != 0 && i < 256 - 1; i++)
		discovery->message[i] = message[i];
	discovery->message[i] = 0;
	discovery->irc_handle = NULL;
	discovery->local_timer = TESTIFY_LOCAL_TIMER + 1.0;
	discovery->local_handle = NULL;
	discovery->local_port_count = 0;
	discovery->global_handle = NULL;
	discovery->global_buffer_used = 0;
	discovery->method_flags = method_flags;
	discovery->mutex = imagine_mutex_create();
//	if(method_flags & TESTIFY_DMF_WAN)
//		discovery->global_handle = testify_discover_global_connect(discovery->service);
	if((method_flags & TESTIFY_DMF_BROADCAST_LOCAL) || (method_flags & TESTIFY_DMF_BROADCAST_LAN) )
		testify_discover_broadcast_initialize(discovery);
	imagine_mutex_lock(testify_status.mutex);
	discovery->next = testify_status.discovery;
	testify_status.discovery = discovery;
	imagine_mutex_unlock(testify_status.mutex);
	return discovery;
}


void testify_discover_destroy(TestifyDiscovery *discover)
{
	TestifyDiscovery **d;
	imagine_mutex_lock(testify_status.mutex);
	for(d = &testify_status.discovery; *d != discover; d = &(*d)->next);
	*d = discover->next;
	imagine_mutex_lock(discover->mutex);
	imagine_mutex_unlock(testify_status.mutex);
	if(discover->local_handle != NULL)
		testify_discover_broadcast_destroy(discover);
	if(discover->peers != NULL)
		free(discover->peers);	
	imagine_mutex_destroy(discover->mutex);	
	free(discover);
}

boolean testify_discover_updated(TestifyDiscovery *discover)
{
	boolean updated;
	imagine_mutex_lock(discover->mutex);
	updated = discover->updated;
	discover->updated = FALSE;
	imagine_mutex_unlock(discover->mutex);
	return updated;
}

TestifyDiscoverPeer	*testify_discover_service_list_get(TestifyDiscovery *discover, uint *peer_count)
{
	TestifyDiscoverPeer	*peers = NULL;
	uint i;
	imagine_mutex_lock(discover->mutex);
	*peer_count = discover->peer_count;
	if(discover->peer_count != 0)
	{
		peers = malloc((sizeof *peers) * discover->peer_count);
		for(i = 0; i < discover->peer_count; i++)
			peers[i] = discover->peers[i].peer;
	}
	imagine_mutex_unlock(discover->mutex);
	return peers;
}