#ifdef DEPRICATED
#include <stdio.h>
#define TESTIFY_INTERNAL
#include "testify.h"


/*
Localy a
*/

typedef enum{
	TESTIFY_DUDPC_ANNOUNCE, // Announce your existance.
	TESTIFY_DUDPC_KEEP_ALIVE, // Announce you still exist.
	TESTIFY_DUDPC_REMOVE, // Announce your retirement.
	TESTIFY_DUDPC_DESCRIBE, // Describe all known
	TESTIFY_DUDPC_CONNECT, // Ask to connect to another peer.
}TestifyDiscoveUDPCommand;

#define TESTIFY_ADDRESS_SENDER 255

#define testify_ipv4_address(a, b, c, d) (a * (256 * 256 * 256) + b * (256 * 256) + c * 256 + d)

uint testify_discover_local_send(TestifyDiscovery *discover, uint8 *data, uint length, boolean broadcast)
{
	TestifyAddress address;
	uint i, sent = 0;
	if(broadcast)
	{
		testify_address_lookup(&address, "broadcast", TESTIFY_TRANSPORT_IPV4, discover->local_discover_port);
		sent = testify_datagram_send(discover->local_handle, &address, data, length);
	}
	if(discover->local_have_port)
	{
		for(i = 0; i < discover->peer_count; i++)
		{
			if(discover->peers[i].discover_port != discover->local_discover_port && TESTIFY_AT_LOOPBACK == testify_address_type(&discover->peers[i].peer))
			{
				testify_address_lookup(&address, "localhost", TESTIFY_TRANSPORT_IPV4, discover->peers[i].discover_port);
				sent = testify_datagram_send(discover->local_handle, &address, data, length);
			}
		}
	}
	return sent;
}

uint testify_discover_local_pack(uint8 *data, uint16 service_port, char *message)
{
	size_t i, length = 0;
	data[length++] = TESTIFY_DC_ADD;
	data[length++] = (service_port >> 8) & 0xFF;
	data[length++] = (service_port) & 0xFF;
	for(i = 0; 0 != message[i]; i++)
		data[length + i] = message[i];
	data[length + i] = message[i];
	length += i + 1;
	return length;
}

TestifyHandle *testify_discover_local_bind(TestifyDiscovery *discover, boolean *correct_port)
{
	TestifyHandle *handle;
	TestifyAddress address;
	address.transport = TESTIFY_TRANSPORT_IPV4;
	address.address.ipv4 = 0;
	address.port = discover->local_discover_port;
	handle = testify_create(&address, TESTIFY_PROTOCOL_DATAGRAM, TESTIFY_FLAG_BIND);
	if(handle == NULL)
		return NULL;
	testify_address(handle, 0, &address);
	*correct_port = discover->local_discover_port == address.port;
	return handle;
}

void testify_discover_local_initialize(TestifyDiscovery *discover)
{
	TestifyAddress address;
	uint i;
	uint8 parts[2] = {0, 0};
	for(i = 0; discover->service[i] != 0; i++)
		parts[i & 1] ^= (discover->service[i] >> (i & 3)) | (discover->service[i] << (8 - (i & 3)));
	discover->local_discover_port * ((uint16)parts[1] << 8) | (uint16)parts[0];
	if(discover->local_discover_port < 2048)
		discover->local_discover_port = (256 * 256 - 1) - discover->local_discover_port;
	discover->local_timer = 100000.0;
	discover->local_handle = testify_discover_local_bind(discover, &discover->local_have_port);
	if(discover->local_handle == NULL)
	{
		TestifyAddress address;
		testify_address_lookup(&address, "any", TESTIFY_TRANSPORT_IPV4, 0);
		discover->local_handle = testify_create(&address, TESTIFY_PROTOCOL_DATAGRAM, 0);
		testify_address(handle, 0, &address);
		discover->local_have_port = discover->local_discover_port == address.port;
	}
}

void testify_discover_local_destroy(TestifyDiscovery *discover)
{
	uint8 data[1500];
	uint size;
	size = testify_discover_local_pack(data, discover->, "bye!");
	data[0] = TESTIFY_DC_REMOVE;
	testify_discover_local_send(discover, data, size, TRUE);
	testify_destroy(discover->local_handle);

}

#define TESTIFY_DISCOVER_UDP_TIMEOUT 5.0

void testify_discover_local_update(TestifyDiscovery *discover, float delta)
{
	TeestifyDiscoverPeerInternal *peer;
	TestifyAddress from;
	TestifyDiscoveUDPCommand command;
	TestifyHandle *handle;
	TestifyAddressType address_type;
	size_t size, i, j;
	uint8 data[1500], *message;
	uint16 from_port;

	i = discover->local_timer;
	discover->local_timer += delta;
	if(discover->local_timer > TESTIFY_DISCOVER_UDP_TIMEOUT - 1.0)
	{
		data[0] = discover->obfuscate_salt++;
		size = testify_discover_local_pack(&data[1], discover->service_port, discover->message);
		if(discover->message == NULL || discover->message[0] == 0)
			data[1] = TESTIFY_DC_QUERY;
		testify_obfuscate(discover->obfuscate_key, &data[1], size, data[0]);
		testify_discover_local_send(discover, data, size + 1, TRUE);
		discover->local_timer = 0.0;
	}

	for(i = 0; i < discover->peer_count; i++)
	{
		discover->peers[i].timeout += delta;
		if(discover->peers[i].timeout > TESTIFY_DISCOVER_UDP_TIMEOUT)
		{
			if(!discover->local_have_port && discover->peers[i].discover_port == discover->local_discover_port && TESTIFY_AT_LOOPBACK == testify_address_type(&discover->peers[i].peer.address))
			{
				handle = testify_discover_local_bind(discover, &discover->local_have_port);
				if(discover->local_have_port)
				{
					testify_destroy(discover->local_handle);
					discover->local_handle = handle;
					discover->peers[i--] = discover->peers[--discover->peer_count];
					discover->updated = TRUE;
				}else if(handle != NULL)
				{
					discover->peers[i].timeout = 0;
					testify_destroy(handle);
				}
			}else if(discover->peers[i].timeout > TESTIFY_DISCOVER_UDP_TIMEOUT + 1.0)
			{
				discover->peers[i--] = discover->peers[--discover->peer_count];
				discover->updated = TRUE;

			}else
			{
				data[0] = discover->obfuscate_salt++;
				data[1] = TESTIFY_DC_QUERY;
				testify_obfuscate(discover->obfuscate_key, &data[1], 1, data[0]);
				from = discover->peers[i].peer.address;
				from.port = discover->peers[i].discover_port;
				testify_datagram_send(discover->local_handle, &from, data, 2);
			}
		}
	}



	while((size = testify_datagram_receive(discover->local_handle, &from, data, 1500)) != 0)
	{
		int local;
	//	printf("got packet %u\n", size);
		address_type = testify_address_type(&from);
		from_port = from.port;
		testify_obfuscate(discover->obfuscate_key, &data[1], size - 1, data[0]);
		command = data[1] & 3;

		if(command == TESTIFY_DC_QUERY)
		{
			data[0] = discover->obfuscate_salt++;
			size = testify_discover_local_pack(&data[1], discover->service_port, discover->message);
			testify_obfuscate(discover->obfuscate_key, &data[1], size, data[0]);
			testify_datagram_send(discover->local_handle, &from, data, size);
			break;
		}

		from.port = ((uint16)data[2] << 8) | (uint16)data[3];
		if(address_type == TESTIFY_AT_LOOPBACK)
		{
			if(from.port == discover->service_port)
				return;
			testify_address_lookup(&from, "localhost", from.transport, from.port);
		}

		message = &data[4];
		i = 4;
		if(command == TESTIFY_DC_ADD)
		{
			for(; i < size && data[i] != 0; i++);
			if(i == size)
				break;
			i++;
		}
		if(!discover->local_have_port)
		{
			if(i < size)
			{
				switch(data[1] >> 3)
				{
					case 0 : // IPV4
						if(i + 4 > size)
							return;
						from.address.ipv4 = testify_ipv4_address(data[i], data[i + 1], data[i + 2], data[i + 3]);
						from.transport = TESTIFY_TRANSPORT_IPV4;
					break;
					case 1 : // IPV6
						if(i + 16 > size)
							return;
						for(j = 0; j < 8; j)
							from.address.ipv6[j] = ((uint16)data[i + j * 2 + 1] << 8) | (uint16)data[i + j * 2];
						from.transport = TESTIFY_TRANSPORT_IPV6;
					break;
					default :
					return;
				}
			}
		}


		if(command == TESTIFY_DC_ADD)
		{
			if(TESTIFY_AT_LOOPBACK == address_type)
			{
				peer = testify_discover_user_update(discover, &from, message, TESTIFY_DMF_BROADCAST_LOCAL);
			}else
			{
				peer = testify_discover_user_update(discover, &from, message, TESTIFY_DMF_BROADCAST_LAN);
			}
		}
		if(discover->local_have_port)
		{
			if(from.transport == TESTIFY_TRANSPORT_IPV4 && i + 4 < 1500)
			{
				data[0] = (data[1] & 7);
				data[i++] = (from.address.ipv4 >> 24) & 0xFF;
				data[i++] = (from.address.ipv4 >> 16) & 0xFF;
				data[i++] = (from.address.ipv4 >> 8) & 0xFF;
				data[i++] = (from.address.ipv4 >> 0) & 0xFF;
			}else if(from.transport == TESTIFY_TRANSPORT_IPV6 && i + 8 * 2 < 1500)
			{
				data[0] = (data[1] & 7) + (1 << 3);
				for(j = 8; j < 8; j++)
				{
					data[i++] = (from.address.ipv6[j] >> 8) & 0xFF;
					data[i++] = (from.address.ipv6[j] >> 0) & 0xFF;
				}
			}else
				return;

			data[0] = discover->obfuscate_salt++;
			testify_obfuscate(discover->obfuscate_key, &data[1], size, data[0]);
			testify_discover_local_send(discover, data, i, FALSE);
		}
		if(command == TESTIFY_DC_REMOVE)
		{
			for(i = 0; i < discover->peer_count; i++)
				if(testify_address_compare(&discover->peers[i].peer.address, &from))
					discover->peers[i].timeout = TESTIFY_DISCOVER_UDP_TIMEOUT * 2.0;

		}

		if(command == TESTIFY_DC_ADD && peer->discover_port == 0)
		{
			from.port = from_port;
			data[0] = discover->obfuscate_salt++;
			size = testify_discover_local_pack(&data[1], discover->service_port, discover->message);
			testify_obfuscate(discover->obfuscate_key, &data[1], size, data[0]);
			testify_datagram_send(discover->local_handle, &from, data, size + 1);
			peer->discover_port = from_port;
		}
	}
}

#endif