/*
Test if a local address could be Static/Public
if(maybe public)
	DNS check.
if(Not static)
{
	-Try PCP
	if(IP address)
		Confirm IP with DNS
}

When connecting:
	look at internal lists of dosovery.

if()

{
	-If(available)  
	{
		-Take IP port.
		-Test DNS
	}
}

-IP
-Hole punch required.
*
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include "forge.h"
#include "imagine.h"
#define TESTIFY_INTERNAL
#include "testify.h"


uint8 *opendns_address = "resolver1.opendns.com";
uint8 opendns_packet[] = {0x26, 0xd4, 0x81, 0x80, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01, 0x04, 0x6d, 0x79, 0x69, 0x70, 0x07, 0x6f, 0x70, 0x65, 0x6e, 0x64, 0x6e, 0x73, 0x03, 0x63, 0x6f, 0x6d, 0x00, 0x00, 0x01, 0x00, 0x01, 0xc0, 0x0c, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x97, 0xec, 0xc9, 0xbd, 0x00, 0x00, 0x29, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

uint8 *google_address = "ns1.google.com";
uint8 google_packet[] = {0xa2, 0x57, 0x01, 0x20, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x03, 0x6f, 0x2d, 0x6f, 0x06, 0x6d, 0x79, 0x61, 0x64, 0x64, 0x72, 0x01, 0x6c, 0x06, 0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 0x03, 0x63, 0x6f, 0x6d, 0x00, 0x00, 0x10, 0x00, 0x01, 0x00, 0x00, 0x29, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

uint8 *cloudflair_address = "1.0.0.1";
uint8 cloudflair_packet[] = {0x25, 0x4d, 0x01, 0x20, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x06, 0x77, 0x68, 0x6f, 0x61, 0x6d, 0x69, 0x0a, 0x63, 0x6c, 0x6f, 0x75, 0x64, 0x66, 0x6c, 0x61, 0x72, 0x65, 0x00, 0x00, 0x10, 0x00, 0x03, 0x00, 0x00, 0x29, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

uint8 *akamai_address = "ns1-1.akamaitech.net";
uint8 akamai_packet[] = {0xeb, 0x9a, 0x01, 0x20, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x06, 0x77, 0x68, 0x6f, 0x61, 0x6d, 0x69, 0x06, 0x61, 0x6b, 0x61, 0x6d, 0x61, 0x69, 0x03, 0x6e, 0x65, 0x74, 0x00, 0x00, 0xff, 0x00, 0x01, 0x00, 0x00, 0x29, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

boolean testify_dns_send_buffer(TestifyHandle *handle, uint service)
{
	TestifyAddress to;
	uint8 *packet = NULL;
	size_t size = 0;
	switch(service % 4)
	{
		case 0 :
			packet = opendns_packet;
			size = sizeof(opendns_packet);
			if(!testify_address_lookup(&to, opendns_address, TESTIFY_TRANSPORT_IPV4, 53))
				return FALSE;
		break;
		case 1 :
			packet = google_packet;
			size = sizeof(google_packet);
			if(!testify_address_lookup(&to, google_address, TESTIFY_TRANSPORT_IPV4, 53))
				return FALSE;
		break;
		case 2 :
			packet = cloudflair_packet;
			size = sizeof(cloudflair_packet);
			if(!testify_address_lookup(&to, cloudflair_address, TESTIFY_TRANSPORT_IPV4, 53))
				return FALSE;
		break;
		case 3 :
			packet = akamai_packet;
			size = sizeof(akamai_packet);
			if(!testify_address_lookup(&to, akamai_address, TESTIFY_TRANSPORT_IPV4, 53))
				return FALSE;
		break;
		default :
		return FALSE;
	}
	return size == testify_datagram_send(handle, &to, packet, size);
}


void testify_dns_send(TestifyHandle *handle, char *dns_name, char *dns_lookup)
{
	 TestifyAddress to;
	uint8 buffer[512] = {'V', 'H', 0x01, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	if(!testify_address_lookup(&to, dns_name, TESTIFY_TRANSPORT_IPV4, 53))
		return;
	uint i, j, pos = 12;
	for(i = 0; TRUE; i += j + 1)
	{
		for(j = 0; dns_lookup[j + i] != 0 && dns_lookup[j + i] != '.'; j++);
		buffer[pos++] = j;
		memcpy(&buffer[pos], &dns_lookup[i], j);
		pos += j;
		if(dns_lookup[i + j] == 0)
			break;
	}
	buffer[pos++] = 0;
	buffer[pos++] = 0;
	buffer[pos++] = 1;
	buffer[pos++] = 0;
	buffer[pos++] = 1;
	testify_datagram_send(handle, &to, buffer, pos);
}


boolean testify_dns_parse_string(uint8 *buffer, uint buffer_size, uint *string)
{
	if(buffer[*string] >= 192)
	{
		*string += 2;
		return TRUE;
	}
	while(buffer[*string] != 0)
	{
		if(*string >= buffer_size)
			return FALSE;
		*string += (uint)buffer[*string] + 1;
	}
	(*string)++;
	return TRUE;
}


boolean testify_dns_parse(uint8 *buffer, uint buffer_size, TestifyAddress *from)
{
	uint i, j, pos = 12, question_count, reply_count, payload_length;
	if(buffer_size < 12/* || buffer[0] != 'V' || buffer[1] != 'H'*/)
		return FALSE; 
	question_count = (uint)buffer[5] + ((uint)buffer[4] << 8);
	reply_count = (uint)buffer[7] + ((uint)buffer[6] << 8);
//	reply_count += (uint)buffer[9] + ((uint)buffer[8] << 8);
//	reply_count += (uint)buffer[11] + ((uint)buffer[10] << 8);

	for(i = 0; i < question_count; i++)
	{
		if(pos + 5 > buffer_size)
			return FALSE;
		if(!testify_dns_parse_string(buffer, buffer_size, &pos))
			return FALSE;
		if(pos + 4 > buffer_size)
			return FALSE;
		pos += 4;
	}

	for(i = 0; i < reply_count; i++)
	{
		if(pos + 5 > buffer_size)
			return FALSE;
		if(!testify_dns_parse_string(buffer, buffer_size, &pos))
			return FALSE;
		if(pos + 2 /* type */ + 2 /* class */ + 4 /* TTL */ + 2 > buffer_size)
			return FALSE;
		pos += 2 /* type */ + 2 /* class */ + 4 /* TTL */;
		payload_length = (uint)buffer[pos + 1] + ((uint)buffer[pos] << 8);
		pos += 2;
		if(pos + payload_length > buffer_size)
			return FALSE;
		if(payload_length == 4)
		{
			from->transport = TESTIFY_TRANSPORT_IPV4;
			from->port = 53;
			from->address.ipv4 = (uint32)buffer[pos + 3] | ((uint32)buffer[pos + 2] << 8) | ((uint32)buffer[pos + 1] << 16) | ((uint32)buffer[pos] << 24);
			return TRUE;
		}else if(payload_length == 8 * 2)
		{
			from->transport = TESTIFY_TRANSPORT_IPV6;
			from->port = 53;
			from->address.ipv6[0] = (uint32)buffer[pos + 1] | ((uint32)buffer[pos] << 8);
			from->address.ipv6[1] = (uint32)buffer[pos + 3] | ((uint32)buffer[pos + 2] << 8);
			from->address.ipv6[2] = (uint32)buffer[pos + 5] | ((uint32)buffer[pos + 4] << 8);
			from->address.ipv6[3] = (uint32)buffer[pos + 7] | ((uint32)buffer[pos + 6] << 8);
			from->address.ipv6[4] = (uint32)buffer[pos + 9] | ((uint32)buffer[pos + 8] << 8);
			from->address.ipv6[5] = (uint32)buffer[pos + 11] | ((uint32)buffer[pos + 10] << 8);
			from->address.ipv6[6] = (uint32)buffer[pos + 13] | ((uint32)buffer[pos + 12] << 8);
			from->address.ipv6[7] = (uint32)buffer[pos + 15] | ((uint32)buffer[pos + 14] << 8);
			return TRUE;
		}
		pos += payload_length;
	}
	return FALSE;
}
