#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "forge.h"
#define TESTIFY_INTERNAL
#include "testify.h"

#ifndef TESTIFY_DEPRICATE_LEGACY_API

#define TESTIFY_MINIMUM_WRITE_SPACE 1024

typedef enum{
	T_TYPE_UINT8,
	T_TYPE_INT8,
	T_TYPE_UINT16,
	T_TYPE_INT16,
	T_TYPE_UINT32,
	T_TYPE_INT32,
	T_TYPE_UINT64,
	T_TYPE_INT64,
	T_TYPE_REAL32,
	T_TYPE_REAL64,
	T_TYPE_RAW,
	T_TYPE_STRING,
	T_TYPE_STRUCT,
	T_TYPE_COUNT
}UTypes;
/*
typedef enum{
	T_HT_STREAMING_SERVER,
	T_HT_STREAMING_CONNECTION,
	T_HT_PACKET_PEER,
	T_HT_FILE_READ,
	T_HT_FILE_WRITE,
	T_HT_BUFFER
}THandleType;

typedef struct{
	TestifyHandle *handle;
	THandleType type;
	FILE *file;
	uint8 *out_buffer;
	size_t out_buffer_allocated;
	size_t out_buffer_used;
	size_t out_buffer_pos;
	uint8 *in_buffer;
	size_t in_buffer_allocated;
	size_t in_buffer_used;
	size_t in_buffer_pos;
	boolean marker;
}THandle;*/

THandle *testify_legacy_handle_allocate(TestifyHandle *handle, THandleType type, uint size)
{
	THandle *legacy_handle;
	legacy_handle = (THandle *)malloc(sizeof *legacy_handle);
	legacy_handle->handle = handle;
	legacy_handle->file = NULL;
	legacy_handle->type = type;
	if(size == 0)
		legacy_handle->out_buffer = NULL;
	else
		legacy_handle->out_buffer = (uint8 *)malloc(size);
	legacy_handle->out_buffer_allocated = size;
	legacy_handle->out_buffer_used = 0;
	if(size == 0)
		legacy_handle->in_buffer = NULL;
	else
		legacy_handle->in_buffer = (uint8 *)malloc(size);
	legacy_handle->in_buffer_allocated = size;
	legacy_handle->in_buffer_used = 0;
	legacy_handle->in_buffer_pos = 0;
	legacy_handle->marker = FALSE;
	return legacy_handle;
}

THandle *testify_network_stream_address_create(const char *host_name, uint16 port)
{
	THandle *legacy_handle;
	TestifyAddress address;
	TestifyHandle *handle;
	if(!testify_address_lookup(&address, host_name, TESTIFY_TRANSPORT_IPV4, port))
		return NULL;
	if(host_name == NULL)
		handle = testify_create(&address, TESTIFY_PROTOCOL_STREAM, TESTIFY_FLAG_BIND);
	else
		handle = testify_create(&address, TESTIFY_PROTOCOL_STREAM, 0);
	if(handle == NULL)
		return NULL;
	return testify_legacy_handle_allocate(handle, T_HT_STREAMING_CONNECTION, 2048);
}


int testify_network_stream_send_force(THandle *handle)
{
	uint i;
	if(handle->handle == NULL)
		return TRUE;
	for(i = 0; testify_connected(handle->handle) && i < handle->out_buffer_used; i += testify_stream_send(handle->handle, &handle->out_buffer[i], handle->out_buffer_used - i));
	handle->out_buffer_used = 0;
	return testify_connected(handle->handle);
}

THandle *testify_network_stream_wait_for_connection(THandle *listen, TestifyNetworkAddress *from)
{
	TestifyHandle *handle;
	handle = testify_wait_for_connection(listen->handle);
	if(handle == NULL)
		return NULL;
	from->ip = handle->address.address.ipv4;
	from->port = handle->address.port;
	return testify_legacy_handle_allocate(handle, T_HT_STREAMING_CONNECTION, 2048);
}

boolean	testify_network_stream_connected(THandle *handle)
{
	return testify_connected(handle->handle);
}

boolean	testify_network_address_lookup(TestifyNetworkAddress *address, char *dns_name, uint16 default_port)
{
	TestifyAddress new_address;
	if(testify_address_lookup(&new_address, dns_name, TESTIFY_TRANSPORT_IPV4, default_port))
	{
		address->ip = new_address.address.ipv4;
		address->port = new_address.port;
		return TRUE;
	}
	return FALSE;
}

THandle *testify_network_datagram_create(uint16 port)
{
	THandle *legacy_handle;
	TestifyAddress address;
	TestifyHandle *handle;
	if(!testify_address_lookup(&address, "any", TESTIFY_TRANSPORT_IPV4, port))
		return NULL;
	handle = testify_create(&address, TESTIFY_PROTOCOL_DATAGRAM, TESTIFY_FLAG_BIND | TESTIFY_FLAG_PORT_MAPPED);
	if(handle == NULL)
		return NULL;
	return testify_legacy_handle_allocate(handle, T_HT_PACKET_PEER, 2048);
}

int testify_network_datagram_send(THandle *handle, TestifyNetworkAddress *to)
{
	TestifyAddress address;
	int out;
	address.transport = TESTIFY_TRANSPORT_IPV4;
	address.address.ipv4 = to->ip;
	address.port = to->port;
	out = (int)testify_datagram_send(handle->handle, &address, handle->out_buffer, handle->out_buffer_used);
	handle->out_buffer_used = 0;
	return out;
}

int testify_network_receive(THandle *handle, TestifyNetworkAddress *from)
{
	TestifyAddress address;
	handle->in_buffer_pos = 0;
	handle->in_buffer_used = testify_datagram_receive(handle->handle, &address, handle->in_buffer, handle->in_buffer_allocated);
	from->ip = address.address.ipv4;
	from->port = address.port;
	return handle->in_buffer_used;
}


boolean testify_network_address_compare(TestifyNetworkAddress *a, TestifyNetworkAddress *b)
{
	return a->ip == b->ip && a->port == b->port;
}

THandle *testify_file_load(char *path) /* Opens a Testify handle from a file name for reading. */
{
	char *debug_header = "UnRaVeLdEbUg";
	uint type_sizes[] = {sizeof(uint8), // T_TYPE_UINT8
									sizeof(int8), // T_TYPE_INT8
									sizeof(uint16), // T_TYPE_UINT16
									sizeof(int16), // T_TYPE_INT16
									sizeof(uint32), // T_TYPE_UINT32
									sizeof(int32), // T_TYPE_INT32
									sizeof(uint64), // T_TYPE_UINT64
									sizeof(int64), // T_TYPE_INT64
									sizeof(float), // T_TYPE_REAL32
									sizeof(double), // T_TYPE_REAL64
									sizeof(uint8)}; // T_TYPE_STRING
	THandle *handle;
	size_t size, i, pos;
	uint32 length;
	uint8 *data, read_type;
	data = (uint8 *)f_text_load(path, &size);
	if(data == NULL)
		return NULL;
	handle = testify_legacy_handle_allocate(NULL, T_HT_FILE_READ, 0);
	handle->in_buffer = data;
	for(i = 0; i < size && debug_header[i] != 0 && data[i] == debug_header[i]; i++);
	if(debug_header[i] == 0)
	{
		pos = 0;
		while(i + 6 < size)
		{
			read_type = data[i++];
			length  = ((uint32) data[i++]) << 24;
			length |= ((uint32) data[i++]) << 16;
			length |= ((uint32) data[i++]) << 8;
			length |= data[i++];
			while(data[i] != 0 && i < size)
				i++;
			i++;
			if(read_type == T_TYPE_STRING)
			{
				while(data[i] != 0 && i < size)
					data[pos++] = data[i++];
				data[pos++] = data[i++];
			}else if(read_type < T_TYPE_STRING)
			{
				memcpy(&data[pos], &data[i], length * type_sizes[read_type]);
				pos += length * type_sizes[read_type];
				i += length * type_sizes[read_type];
			}else
			{
				printf("Testify Error: Parsing of file with debug data failed!\n");
				exit(0);
			}
		}
		handle->in_buffer_used = pos;
	}else
		handle->in_buffer_used = size;
	handle->in_buffer_allocated = handle->in_buffer_used = size;
	return handle;
}


THandle *testify_file_save(char *path)
{
	THandle *handle;
	FILE *f;
	f = fopen(path, "wb");
	if(f == NULL)
		return NULL;
	handle = testify_legacy_handle_allocate(NULL, T_HT_FILE_WRITE, 0);
	handle->file = f;
	return handle;
}

/* ---------- Memmory buffers --------
Handles the creation and mangement of in memory FIFO channels.*/

THandle *testify_buffer_create()
{
	return testify_legacy_handle_allocate(NULL, T_HT_BUFFER, 1024);
}

void *testify_buffer_get(THandle *handle, uint32 *size)
{
	*size = handle->out_buffer_used;
	return handle->out_buffer;
}

void testify_buffer_set(THandle *handle, void *data, uint32 size) /*Lets the user give a pointer to "data" of "Size" number of bytes that will be added to the buffer. */
{

	if(handle->out_buffer_allocated - handle->out_buffer_used < size)
	{
		handle->out_buffer_allocated = handle->out_buffer_used + size;
		handle->out_buffer = realloc(handle->out_buffer, handle->out_buffer_allocated);
	}
	memcpy(&handle->out_buffer[handle->out_buffer_used], data, size);
	handle->out_buffer_used += size;
}

/* ---------- Handle Utillities --------
*/

THandleType	testify_type(THandle *handle)
{
	return handle->type;
}

void testify_free(THandle *handle)
{
	if(handle->file != NULL)
	{
		fwrite(handle->out_buffer, 1, handle->out_buffer_used, handle->file);
		fclose(handle->file);
	}
	free(handle);
}

extern uint		testify_network_wait(THandle **handles, boolean *read, boolean *write, unsigned int handle_count, unsigned int microseconds); /* querys if network handles are ready to read and write. Can handle multiple handles as speccified by the array "handles" and handle count. The function will block for at most milliseconds and wait for any resource to become availbale. Once a resource gets available it will return. */

/* ---------- Debugging --------
Testify has an extencive debugging iunfrastructure desinged to enable inrosprection of binary data stream and the abillity to debug miss matches of how a */

void testify_debug_mode_set(THandle *handle, boolean debug, char *text_copy_name)
{

}

void testify_network_debug_incomming(THandle *handle)
{
	printf("Testify: incomming data stream:\n");
	f_print_raw(handle->in_buffer, handle->in_buffer_used);
}

void testify_network_debug_outbound(THandle *handle)
{
	printf("Testify: outbound data stream:\n");
	f_print_raw(handle->out_buffer, handle->out_buffer_used);
}

void testify_network_debug_address_print(TestifyNetworkAddress *address, char *buffer)
{
	sprintf(buffer, "%u.%u.%u.%u:%u", (address->ip >> 24) & 0xFF, (address->ip >> 16) & 0xFF, (address->ip >> 8) & 0xFF, address->ip & 0xFF,  address->port);
}


/* ---------- Markers --------
Markers allow a application to set down a marker, then parse forward, and then if need be revert to the marker position in case not enought data has arrived to complete the parse. Markers are especialy imporatnt whjen dealing with network handles, but it is always good practice to use them to make sure that the data one tries to retrive is intact.
*/

void testify_unpack_colapse(THandle *handle)
{
	size_t i, length, gap;
	gap = handle->in_buffer_pos;
	length = handle->in_buffer_used - handle->in_buffer_pos;
	for(i = 0; i < length; i++)
		handle->in_buffer[i] = handle->in_buffer[i + gap];
	handle->in_buffer_pos = 0;
	handle->in_buffer_used = length;
}

void testify_restart_marker_set(THandle *handle)
{
	size_t i, length, gap;
	if(handle->in_buffer_pos != 0)
		testify_unpack_colapse(handle);
	handle->marker = TRUE;
}

void testify_restart_marker_release(THandle *handle)
{
	handle->marker = FALSE;
}

void testify_restart_marker_reset(THandle *handle)
{
	handle->in_buffer_pos = 0;
}


void testify_unpack_prepare(THandle *handle, size_t size)
{
	uint8 *b;
	uint i, length;
	if(handle->in_buffer_used - handle->in_buffer_pos < size)
	{
		if(handle->in_buffer_used + size > handle->in_buffer_allocated)
		{
			if(handle->in_buffer_used - handle->in_buffer_pos + size < handle->in_buffer_allocated)
			{
				testify_unpack_colapse(handle);
			}else
			{
				b = realloc(handle->in_buffer, handle->in_buffer_used + size + 1024);
				if(b != NULL)
				{
					handle->in_buffer = b;
					handle->in_buffer_allocated = handle->in_buffer_used + size + 1024;
				}
			}
		}
		if(handle->type == T_HT_STREAMING_SERVER || handle->type == T_HT_STREAMING_CONNECTION)
			handle->in_buffer_used += testify_stream_recive(handle->handle, &handle->in_buffer[handle->in_buffer_used], handle->in_buffer_allocated - handle->in_buffer_used);
	}
}

boolean	testify_retivable(THandle *handle, uint size)
{
	testify_unpack_prepare(handle, size);


	return handle->in_buffer_used - handle->in_buffer_pos >= size;
}

boolean	testify_retivable_terminated(THandle *handle)
{
	uint i;
	for(i = handle->in_buffer_pos; i < handle->in_buffer_used; i++)
		if(handle->in_buffer[i] == 0)
			return TRUE;
	testify_unpack_prepare(handle, 64);
	return FALSE;
}
/* ---------- Data packing --------
Functions used to pack data in to a stram. If TESTIFY_DEBUG is not defined, the param "name" will be removed with a macro. In debug mode (where both TESTIFY_DEBUG is defined AND the bug mode of the stream has been set to TRUE testify_debug_mode_set) the name, and data type will be checked to make sure they match the name and type that was packed. If this test fails the application will terminate on unpack and write out a Error raport to standard out.*/

void testify_pack_prepare(THandle *handle, size_t size)
{
	if(handle->out_buffer_used + size > handle->out_buffer_allocated)
	{
		switch(handle->type)
		{
			case T_HT_STREAMING_SERVER :
			printf("Testify Error: Wrapper tries to pack in to a UDP packet beyond the MTU.\n");
			exit(0);
			case T_HT_STREAMING_CONNECTION :
			testify_network_stream_send_force(handle);
			break;
			case T_HT_PACKET_PEER :
			printf("Testify Error: Wrapper tries to pack in to a UDP packet beyond the MTU.\n");
			exit(0);
			case T_HT_FILE_READ :
			printf("Testify Error: Wrapper tries to pack in to a read file\n");
			exit(0);
			case T_HT_FILE_WRITE :
			fwrite(handle->out_buffer, 1, handle->out_buffer_used, handle->file);
			handle->out_buffer_used = 0;
			if(size < handle->out_buffer_allocated)
				return;
			break;
			case T_HT_BUFFER :
			break;
		}
		if(handle->out_buffer_allocated + 4069 > handle->out_buffer_used + size)
			handle->out_buffer_allocated += 4069;
		else
			handle->out_buffer_allocated = handle->out_buffer_used + size;
		handle->out_buffer = realloc(handle->out_buffer, handle->out_buffer_allocated);
	}
}

void testify_pack_uint8(THandle *handle, uint8 value, char *name)
{
	testify_pack_prepare(handle, 1);
	handle->out_buffer[handle->out_buffer_used++] = value;
}

void testify_pack_int8(THandle *handle, int8 value, char *name)
{
	testify_pack_prepare(handle, 1);
	handle->out_buffer[handle->out_buffer_used++] = value;
}

void testify_pack_uint16(THandle *handle, uint16 value, char *name)
{
	testify_pack_prepare(handle, sizeof(uint16));
	testify_pack_bigendian16(&handle->out_buffer[handle->out_buffer_used], (uint8 *)&value);
	handle->out_buffer_used += sizeof(uint16);
}

void testify_pack_int16(THandle *handle, int16 value, char *name)
{
	testify_pack_prepare(handle, sizeof(int16));
	testify_pack_bigendian16(&handle->out_buffer[handle->out_buffer_used], (uint8 *)&value);
	handle->out_buffer_used += sizeof(int16);
}

void testify_pack_uint32(THandle *handle, uint32 value, char *name)
{
	testify_pack_prepare(handle, sizeof(uint32));
	testify_pack_bigendian32(&handle->out_buffer[handle->out_buffer_used], (uint8 *)&value);
	handle->out_buffer_used += sizeof(uint32);
}

void testify_pack_int32(THandle *handle, int32 value, char *name)
{
	testify_pack_prepare(handle, sizeof(int32));
	testify_pack_bigendian32(&handle->out_buffer[handle->out_buffer_used], (uint8 *)&value);
	handle->out_buffer_used += sizeof(uint32);
}

void testify_pack_uint64(THandle *handle, uint64 value, char *name)
{
	testify_pack_prepare(handle, sizeof(uint64));
	testify_pack_bigendian64(&handle->out_buffer[handle->out_buffer_used], (uint8 *)&value);
	handle->out_buffer_used += sizeof(uint64);
}

void testify_pack_int64(THandle *handle, int64 value, char *name)
{
	testify_pack_prepare(handle, sizeof(int64));
	testify_pack_bigendian64(&handle->out_buffer[handle->out_buffer_used], (uint8 *)&value);
	handle->out_buffer_used += sizeof(uint64);
}

void testify_pack_real32(THandle *handle, float value, char *name)
{
	testify_pack_prepare(handle, sizeof(float));
	testify_pack_bigendian32(&handle->out_buffer[handle->out_buffer_used], (uint8 *)&value);
	handle->out_buffer_used += sizeof(float);
}

void testify_pack_real64(THandle *handle, double value, char *name)
{
	testify_pack_prepare(handle, sizeof(uint64));
	testify_pack_bigendian64(&handle->out_buffer[handle->out_buffer_used], (uint8 *)&value);
	handle->out_buffer_used += sizeof(uint64);
}

uint64 testify_pack_raw(THandle *handle, uint8 *data, uint64 length, char *name)
{
	testify_pack_prepare(handle, length);
	memcpy(&handle->out_buffer[handle->out_buffer_used], data, length);
	handle->out_buffer_used += length;
	return length;
}

boolean testify_pack_string(THandle *handle, char *value, char *name)
{
	size_t i;
	for(i = 0; value[i] != 0; i++);
	i++;
	testify_pack_prepare(handle, i);
	testify_pack_raw(handle, value, i, NULL);
	return TRUE;
}


/* ---------- Data unpacking --------
Functions used to unpack data in to a stram. If TESTIFY_DEBUG is not defined, the param "name" will be removed with a macro. In debug mode (where both TESTIFY_DEBUG is defined AND the bug mode of the stream has been set to TRUE testify_debug_mode_set) the name, and data type will be checked to make sure they match the name and type that was packed. If this test fails the application will terminate and write out a Error raport to standard out.*/

uint8 testify_unpack_uint8(THandle *handle, char *name)
{
	return handle->in_buffer[handle->in_buffer_pos++];
}

int8 testify_unpack_int8(THandle *handle, char *name)
{
	return handle->in_buffer[handle->in_buffer_pos++];
}

uint16 testify_unpack_uint16(THandle *handle, char *name)
{
	uint16 value;
	testify_pack_bigendian16((uint8 *)&value, &handle->in_buffer[handle->in_buffer_pos]);
	handle->in_buffer_pos += sizeof(uint16);
	return value;
}

int16 testify_unpack_int16(THandle *handle, char *name)
{
	int16 value;
	testify_pack_bigendian16((uint8 *)&value, &handle->in_buffer[handle->in_buffer_pos]);
	handle->in_buffer_pos += sizeof(int16);
	return value;
}


uint32 testify_unpack_uint32(THandle *handle, char *name)
{
	uint32 value;
	testify_pack_bigendian32((uint8 *)&value, &handle->in_buffer[handle->in_buffer_pos]);
	handle->in_buffer_pos += sizeof(uint32);
	return value;
}

int32 testify_unpack_int32(THandle *handle, char *name)
{
	int32 value;
	testify_pack_bigendian32((uint8 *)&value, &handle->in_buffer[handle->in_buffer_pos]);
	handle->in_buffer_pos += sizeof(int32);
	return value;
}

uint64 testify_unpack_uint64(THandle *handle, char *name)
{
	int64 value;
	testify_pack_bigendian64((uint8 *)&value, &handle->in_buffer[handle->in_buffer_pos]);
	handle->in_buffer_pos += sizeof(int64);
	return value;
}

extern int64 testify_unpack_int64(THandle *handle, char *name)
{
	int64 value;
	testify_pack_bigendian64((uint8 *)&value, &handle->in_buffer[handle->in_buffer_pos]);
	handle->in_buffer_pos += sizeof(int64);
	return value;
}

float testify_unpack_real32(THandle *handle, char *name)
{
	float value;
	testify_pack_bigendian32((uint8 *)&value, &handle->in_buffer[handle->in_buffer_pos]);
	handle->in_buffer_pos += sizeof(float);
	return value;
}

double testify_unpack_real64(THandle *handle, char *name)
{
	double value;
	testify_pack_bigendian64((uint8 *)&value, &handle->in_buffer[handle->in_buffer_pos]);
	handle->in_buffer_pos += sizeof(double);
	return value;
}

boolean testify_unpack_string(THandle *handle, char *value, uint buffer_size, char *name)
{
	uint i;
	buffer_size--;
	for(i = 0; i < buffer_size && handle->in_buffer[handle->in_buffer_pos] != 0; i++)
		value[i] = handle->in_buffer[handle->in_buffer_pos++];
	value[i] = 0;
	while(handle->in_buffer[handle->in_buffer_pos] != 0)
		handle->in_buffer_pos++;
	handle->in_buffer_pos++;
	return TRUE;
}

char *testify_unpack_string_allocate(THandle *handle, char *name)
{
	char *buffer;
	uint length;
	for(length = 0; handle->in_buffer[handle->in_buffer_pos + length] != 0; length++);
	length++;
	buffer = malloc(length);
	memcpy(buffer, &handle->in_buffer[handle->in_buffer_pos], length);
	handle->in_buffer_pos += length;
	return buffer;
}

real64	testify_unpack_raw(THandle *handle, uint8 *buffer, uint64 buffer_length, char *name)
{
	testify_unpack_prepare(handle, buffer_length);
	if(handle->in_buffer_used != 0)
	{
		if(buffer_length >= handle->in_buffer_used - handle->in_buffer_pos)
		{
			buffer_length = handle->in_buffer_used - handle->in_buffer_pos;
			memcpy(buffer, &handle->in_buffer[handle->in_buffer_pos], buffer_length);
			handle->in_buffer_pos = 0;
			handle->in_buffer_used = 0;
		}else
		{
			memcpy(buffer, &handle->in_buffer[handle->in_buffer_pos], buffer_length);
			handle->in_buffer_pos += buffer_length;
		}
		return buffer_length;
	}
	return 0;
}

#endif 