
#define TESTIFY_INTERNAL
#define TESTIFY_MINIMUM_WRITE_SPACE 1024
#define TESTIFY_DISCOVER_UDP_TIMEOUT 5.0

typedef struct{
	char user_name[64];
	char message[512 - 64];
}TestifyIRCMessage;

typedef struct{
	uint flags;
	char name[64];
}TestifyIRCUser;

typedef struct{
	uint8 name[64];
	TestifyIRCUser *users;
	uint user_count;
	uint user_allocated;
	TestifyIRCMessage *messages;
	uint messages_buffer_used;
	uint messages_buffer_pos;
}TestifyIRCChannel;

typedef struct{
	uint8 *buffer;
	size_t buffer_used;
	size_t buffer_allocated;
	TestifyIRCChannel *channels;
	uint channel_count;
	uint message_buffer_size;
	boolean accepted;
	boolean connected;
	char channel_auto_join[64];
	char nick[24];
	uint version;
}TestifyIRC;

typedef struct{
	TestifyAddress address;
	TestifyAddress gateway;
	uint timeout;
}TestifyAddressMapping;

typedef struct{
	int socket;
	uint32 flags;
	boolean connected;
	TestifyProtocol protocol;
	TestifyAddress address;
	TestifyAddressMapping mappings[TESTIFY_HANDLE_ADDRESS_COUNT_MAX];
	uint mapping_count;
	boolean bound_to_specific_ip;
	void *next_mapped_handle;
	void *custom;
}TestifyHandle;

#define TESTIFY_DISCOVER_USER_ID_LENGTH 6

typedef struct{
	TestifyDiscoverPeer peer;
	TestifyAddress address_local;
	TestifyAddress address_global;
	uint16 discover_port;
	uint8 user_id[TESTIFY_DISCOVER_USER_ID_LENGTH];
	double timeout;
}TeestifyDiscoverPeerInternal;


#define TESTIFY_DISCOVER_GLOBAL_BUFFER_SIZE 512


typedef enum{
	TESTIFY_DC_ADD,
	TESTIFY_DC_QUERY,
	TESTIFY_DC_REMOVE,
	TESTIFY_DC_COUNT
}TestifyDiscoverCommand;

#define TESTIFY_LOCAL_PORT_COUNT_MAX 32

typedef struct{
	TeestifyDiscoverPeerInternal *peers;	
	TestifyHandle	*service_handle;
	uint			peer_count;
	uint			peer_allocted;
	uint			peer_start;
	uint32			method_flags;
	TestifyHandle	*local_handle;
	uint16			local_discover_port;
	boolean			local_have_port;
	uint			local_port_count;
					struct{
						uint16 port;
						uint16 timer;
					}local_ports[TESTIFY_LOCAL_PORT_COUNT_MAX];
	double			local_timer;
	TestifyHandle	*global_handle;
	uint8			global_buffer[TESTIFY_DISCOVER_GLOBAL_BUFFER_SIZE];
	uint32			global_buffer_used;
	TestifyHandle	*irc_handle;
	char			service[64];
	char			message[256];
	boolean			updated;
	uint8			obfuscate_salt;
	uint8			obfuscate_key[8];
	void			*mutex;
	void			*next;
}TestifyDiscovery;


#define TESTIFY_PORTMAPPING_TIMER 10.0
#define TESTIFY_PORTMAPPING_IP_COUNT 4
#define TESTIFY_PORTMAPPING_DNS_FAST_TIMER 10.0
#define TESTIFY_PORTMAPPING_DNS_SLOW_TIMER 300.0


typedef struct{
	TestifyServiceType type;
	char *name;
	TestifyAddress address;
	TestifyServiceState status;
	float last_updated;
	float first_detected;
	TestifyHandle *handle;
}TestifyStatusService;

typedef struct{
	TestifyStatusService *services;
	uint service_count;
	uint service_allocated;
	TestifyHandle *mapped_ports;
	uint64 status_bits;
	TestifyHandle *handle;
	TestifyHandle *irc_handle;
	uint irc_version;
	TestifyDiscovery *discovery;
	void *mutex;
}TestifyStatus;

#define testify_ipv4_address(a, b, c, d) (a * (256 * 256 * 256) + b * (256 * 256) + c * 256 + d)

/*
typedef struct{
	uint8 user_id[TESTIFY_DISCOVER_USER_ID_LENGTH];
	uint16 relay_port;
	double timeout;
}TeestifyDiscoverUser;

typedef struct{
	void *mutex;
	TestifyDiscoverPeer *peers_public;
	TeestifyDiscoverUser *peers_private;
	uint peer_count;
	uint peer_allocated;
	char *service;
	char channel[64];
	char message[200];
	uint16 port;
	uint16 ref_count;
	boolean joined;
	boolean global;
	boolean local;
	boolean updated;
	double udp_ping_timer;
	void *next;
}TestifyDiscovery;
*/

extern TestifyStatus testify_status;

/* t_portmap.c */

extern void testify_portmap_update(TestifyStatus *status);
extern void testify_portmap_recive(TestifyStatus *status, TestifyAddress *from, uint8 *buffer, uint size);

extern TeestifyDiscoverPeerInternal *testify_discover_peer_lookup(TestifyDiscovery *discovery, TestifyAddress *address, boolean local);
extern TeestifyDiscoverPeerInternal *testify_discover_peer_update(TestifyDiscovery *discovery, TestifyAddress *address_local, TestifyAddress *address_global, char *message, uint type, uint timeout);
extern void testify_discover_peer_delete(TestifyDiscovery *discovery,  TeestifyDiscoverPeerInternal *peer);


extern void testify_discover_broadcast_initialize(TestifyDiscovery *discover);
extern void testify_discover_broadcast_update(TestifyDiscovery *discover, float delta);
extern void testify_discover_broadcast_destroy(TestifyDiscovery *discover);

extern boolean testify_service_is_localhost(TestifyAddress *address);

extern uint64 testify_status_flags_get();
extern void testify_status_flags_set(uint64 flags, uint64 mask);
extern uint testify_status_service_select(TestifyServiceType type, char *name, uint buffer_size, TestifyAddress *address, TestifyServiceState status);
extern void testify_status_service_status_set(TestifyServiceType type, char *name, TestifyServiceState status);


/* t_discover_irc.c */
extern void testify_discover_update(TestifyDiscovery *discover, double delta);
extern void testify_irc_destroy(TestifyIRC *irc);
extern void testify_discover_irc_command_func(uint command, char *sender, char *channel, char *mesassage, void *user);
TestifyHandle *testify_discover_irc_connect(TestifyDiscovery *discover, char *dns, TestifyAddress *address, uint8 *service);
extern void testify_obfuscate(uint8 *key, uint8 *data, uint length, uint8 salt);

#include "testify.h"