
#include <string.h>
#include "forge.h"
#include "imagine.h"
#define TESTIFY_INTERNAL
#include "testify.h"

void testify_irc_channel_add_users(TestifyIRCChannel *channel, char *names)
{
	TestifyIRCUser *users;
	uint i, j, k, allocated;
	if(channel == NULL)
		return;
	for(i = 0; names[i] != 0; i++)
	{
		if(names[i] > ' ')
		{
			for(j = 0; j < channel->user_count; j++)
			{
				for(k = 0; channel->users[j].name[k] == names[i + k] && channel->users[j].name[k] != 0; k++);
				if(channel->users[j].name[k] == 0 && names[i + k] <= ' ')
					break;
			}
			if(j < channel->user_count)
			{			
				for(j = 0; j < 64 && names[i + j] > ' '; j++);
			}else
			{
				if(channel->user_count == channel->user_allocated)
				{
					if(channel->user_allocated == 0)
						allocated = 32;
					else
						allocated = channel->user_allocated * 2;
					users = realloc(channel->users, (sizeof *channel->users) * allocated);
					if(users != NULL)
					{
						channel->user_allocated = allocated;
						channel->users = users;
					}else
						return;
				}
				channel->users[channel->user_count].flags = 0;
				for(j = 0; j < 64 && names[i + j] > ' '; j++)
					channel->users[channel->user_count].name[j] = names[i + j];
				channel->users[channel->user_count++].name[j] = 0;
			}
			i += j - 1;
		}
	}
}

void testify_irc_channel_remove_users(TestifyIRCChannel *channel, char *names)
{
	uint i, j;
	for(i = 0; i < channel->user_count; i++)
	{
		for(j = 0; channel->users[i].name[j] == names[j] && channel->users[i].name[j] != 0; j++);
		if(channel->users[i].name[j] == 0 && names[j] <= ' ')
		{	
			channel->users[i] = channel->users[--channel->user_count];
			return;
		}
	}
}


uint testify_irc_channel_count_get(TestifyHandle *handle)
{
	TestifyIRC *irc;
	irc = handle->custom;
	return irc->channel_count;
}

char *testify_irc_channel_name_get(TestifyHandle *handle, uint channel)
{
	TestifyIRC *irc;
	irc = handle->custom;
	return irc->channels[channel].name;
}

uint testify_irc_channel_user_count_get(TestifyHandle *handle, uint channel)
{
	TestifyIRC *irc;
	irc = handle->custom;
	return irc->channels[channel].user_count;
}


char *testify_irc_channel_user_name_get(TestifyHandle *handle, uint channel, uint user_id)
{
	TestifyIRC *irc;
	irc = handle->custom;
	return irc->channels[channel].users[user_id].name;
}

uint testify_irc_channel_message_count_get(TestifyHandle *handle, uint channel)
{
	TestifyIRC *irc;
	irc = handle->custom;
	return irc->channels[channel].messages_buffer_used;
}

char *testify_irc_channel_message_get(TestifyHandle *handle, uint channel, uint message_id)
{
	TestifyIRC *irc;
	irc = handle->custom;
	return irc->channels[channel].messages[(irc->channels[channel].messages_buffer_pos + message_id) % irc->message_buffer_size].message;
}

char *testify_irc_channel_message_user_name_get(TestifyHandle *handle, uint channel, uint message_id)
{
	TestifyIRC *irc;
	irc = handle->custom;
	return irc->channels[channel].messages[(irc->channels[channel].messages_buffer_pos + message_id) % irc->message_buffer_size].user_name;
}


void *testify_irc_channel_get(TestifyHandle *handle, char *channel)
{
	uint i, j;
	TestifyIRC *irc;
	if(channel == NULL)
		return NULL;
	irc = handle->custom;
	for(i = 0; i < irc->channel_count; i++)
	{
		for(j = 0; channel[j] != 0 && j < 64 - 1 && channel[j] == irc->channels[i].name[j]; j++);
		if(channel[j] == irc->channels[i].name[j])
			return &irc->channels[i];
		
	}
	return NULL;
}

TestifyIRCChannel *testify_irc_channel_greate(TestifyHandle *handle, char *channel_name)
{
	TestifyIRCChannel *channel;
	uint i, j;
	TestifyIRC *irc;
	irc = handle->custom;
	if(irc->channel_count % 8 == 0)
	{
		irc->channels = realloc(irc->channels, (sizeof *irc->channels) * (irc->channel_count + 8));
	}
	channel = &irc->channels[irc->channel_count++];
	for(i = 0; channel_name[i] != 0 && i < 64 - 1; i++)
		channel->name[i] = channel_name[i];
	channel->name[i] = 0;
	channel->users = NULL;
	channel->user_count = 0;
	channel->user_allocated = 0;
	channel->messages = malloc((sizeof *channel->messages) * irc->message_buffer_size);
	channel->messages_buffer_used = 0;
	channel->messages_buffer_pos = 0;
	return channel;
}

void testify_irc_channel_join(TestifyHandle *handle, char *channel_name)
{
	char buffer[1500];
	uint i;
	sprintf(buffer, "JOIN #%s\r\n", channel_name);
//	printf("%s", buffer);
	for(i = 0; buffer[i] != 0; i++);
	testify_stream_send(handle, buffer, i);
}


TestifyHandle *testify_irc_create(char *server, char *user_name, char *channel)
{
	TestifyIRC *irc;
	TestifyAddress address;
	TestifyHandle *handle = NULL;
	char *read, rand[6], message[512];
	uint i;
	if(testify_address_lookup(&address, server, TESTIFY_TRANSPORT_IPV4, 6667) ||
		testify_address_lookup(&address, server, TESTIFY_TRANSPORT_IPV6, 6667))
		handle = testify_create(&address, TESTIFY_PROTOCOL_STREAM, 0);
	if(handle != NULL)
	{

		handle->custom = irc = malloc(sizeof *irc);
		if(channel != NULL)
		{
			for(i = 0; i < 64 - 1 && channel[i] != 0; i++)
				irc->channel_auto_join[i] = channel[i];
			irc->channel_auto_join[i] = 0;
		}else
			irc->channel_auto_join[0] = 0;
		for(i = 0; i < 24 - 1 && user_name[i] != 0; i++)
			irc->nick[i] = user_name[i];
		irc->nick[i] = 0;

		irc->connected = FALSE;
		irc->buffer_allocated = 1500;
		irc->buffer_used = 0;
		irc->buffer = malloc(irc->buffer_allocated);
		irc->channels = NULL;
		irc->channel_count = 0;
		irc->message_buffer_size = 256;
	}
	return handle;
}

void testify_irc_destroy(TestifyIRC *irc)
{
	free(irc->buffer);
	free(irc);
}


void testify_irc_message_add(TestifyHandle *handle, TestifyIRCChannel *channel_struct, char *name, char *payload)	
{
	TestifyIRC *irc;
	uint i;
	if(channel_struct == NULL)
		return;
	irc = handle->custom;
	channel_struct->messages_buffer_pos = (channel_struct->messages_buffer_pos + irc->message_buffer_size - 1) % irc->message_buffer_size;
	if(channel_struct->messages_buffer_used < irc->message_buffer_size)
		channel_struct->messages_buffer_used++;
	for(i = 0; i < 63 && name[i] != 0; i++)
		channel_struct->messages[channel_struct->messages_buffer_pos].user_name[i] = name[i];
	channel_struct->messages[channel_struct->messages_buffer_pos].user_name[i] = 0;
	for(i = 0; i < 63 && payload[i] != 0; i++)
		channel_struct->messages[channel_struct->messages_buffer_pos].message[i] = payload[i];
	channel_struct->messages[channel_struct->messages_buffer_pos].message[i] = 0;
}


void testify_irc_channel_message_send(TestifyHandle *handle, uint channel, char *message)
{
	TestifyIRC *irc;
	char buffer[1500];
	uint i;
	irc = handle->custom;
	sprintf(buffer, "PRIVMSG #%s :%s\r\n", irc->channels[channel].name, message);
	for(i = 0; buffer[i] != 0; i++);
	testify_stream_send(handle, buffer, i);
	testify_irc_message_add(handle, &irc->channels[channel], "you", message);
}



void testify_irc_buffer_parser(char *buffer, char **command, char **payload, char **name, char **channel)
{
	uint8 *read;
	uint i;
	read = buffer;
	if(read[0] == ':')
	{
		while(*read > ' ')
			read++;
		while(*read <= ' ' && *read != 0)
			read++;
	}
	*command = read;
	while(*read != 0 && *read != ':')
		read++;
	if(*read != 0)
		*payload = ++read;
	else
		*payload = NULL;
	*channel = *name = NULL; 
	if(buffer[0] == ':')
	{
		for(i = 1; buffer[i] != '#' && buffer[i] != 0; i++);
		if(buffer[i] == '#')
		{
			*channel = &buffer[i + 1];
			while(buffer[i] > ' ')
				i++;
			buffer[i] = 0;
		}

		for(i = 1; buffer[i] != '!' && buffer[i] != 0; i++);
		if(buffer[i] == '!')
		{
			buffer[i] = 0;
			*name = &buffer[1];
		}
	}
	read = *command;
	for(i = 0; read[i] > ' '; i++);
	read[i] = 0;
}


uint testify_irc_command_parser(char *command)
{
	char *known_commands[] = {"JOIN", "PART", "PRIVMSG", "NOTICE", "NICK", "NAMES", "PING", "QUIT", "SERVLIST", "TOPIC", "USERS", "OPER"};
	uint i, j;
	if(command[0] >= '0' && command[0] <= '9')
	{
		return (command[0] - '0') * 100 + (command[1] - '0') * 10 + (command[2] - '0');
	}else
	{
		for(i = 0; i < 12; i++)
		{
			for(j = 0; known_commands[i][j] == command[j] && command[j] != 0; j++);
			if(known_commands[i][j] == command[j])
				return TESTIFY_IRC_COMMAND_JOIN + i;
		}
	//	printf("Unknown command: %s\n", command);
	}
	return ~0;
}

void testify_irc_nick_set(TestifyHandle *handle, char *nick)
{
	TestifyIRC *irc;
	char message[256];
	uint i;
	irc = handle->custom;
	for(i = 0; nick[i] != 0; i++);
		irc->nick[i] = nick[i];
	irc->nick[i] = 0;
	for(i = 0; message[i] != 0; i++);
	sprintf(message, "NICK %s\r\n", irc->nick);
//	printf("sending:%s", message);
	for(i = 0; message[i] != 0; i++);
	testify_stream_send(handle, message, i);
}

char *testify_irc_nick_get(TestifyHandle *handle)
{
	TestifyIRC *irc;
	irc = handle->custom;
	return irc->nick;
}

TestifyHandle *testify_irc_parser(TestifyHandle *handle, void (*command_func)(uint command, char *sender, char *channel, char *mesassage, void *user), void *user)
{
	char buffer[1500];
	TestifyIRCChannel *channel_struct;
	TestifyIRC *irc;
	size_t i, j, gap;
	uint op;
	if(handle == NULL)
		return NULL;
	irc = handle->custom;
	if(!irc->connected)
	{
		char message[256];
		irc->connected = TRUE;

		sprintf(message, "NICK %s\r\n", irc->nick);
	//	printf("sending:%s", message);
		for(i = 0; message[i] != 0; i++);
		testify_stream_send(handle, message, i);
		sprintf(message, "USER x * 0 : %s\r\n", irc->nick);
	//	printf("sending:%s", message);
		for(i = 0; message[i] != 0; i++);
		testify_stream_send(handle, message, i);
		if(irc->channel_auto_join[0] != 0)
		{
			sprintf(message, "JOIN #%s\r\n", irc->channel_auto_join);
		//	printf("%s", message);
			for(i = 0; message[i] != 0; i++);
			testify_stream_send(handle, message, i);
		}
	}
	while(TRUE)	
	{
		if(irc->buffer_used == irc->buffer_allocated)	
		{
			if(irc->buffer_used > 10000)
			{
				testify_destroy(handle);
				return NULL;
			}
			irc->buffer_allocated *= 2;
			irc->buffer = realloc(irc->buffer, irc->buffer_allocated);
		}	
		irc->buffer_used += testify_stream_recive(handle, &irc->buffer[irc->buffer_used], irc->buffer_allocated - irc->buffer_used);
		if(!testify_connected(handle))
		{
			testify_destroy(handle);
			return NULL;
		}		
		if(irc->buffer_used == 0)
			return handle;
		for(i = 1; i < irc->buffer_used && (irc->buffer[i] != '\n'  || irc->buffer[i - 1] != '\r'); i++);
		if(i < irc->buffer_used)
		{
			uint8 *read, *command, *payload, *name, *channel;
			irc->buffer[i - 1] = 0;
			gap = i + 1;
			
		//	printf("%s\n", irc->buffer);
			testify_irc_buffer_parser(irc->buffer, &command, &payload, &name, &channel);
			op = testify_irc_command_parser(command);
		//	if(op == ~0)
		//		printf("Unknown command: %s\n", command);
			switch(op)
			{
				case 001 :
				//	printf("We have joined!\n");
					irc->accepted = TRUE;
					if(irc->channel_auto_join[0] != 0)
						testify_irc_channel_join(handle, irc->channel_auto_join);
				break;
				case TESTIFY_IRC_COMMAND_PING :
				{				
					sprintf(buffer, "PONG %s\r\n", payload);
					for(i = 0; buffer[i] != 0; i++);
					testify_stream_send(handle, buffer, i);
				}
				break;
				case 353 :
					testify_irc_channel_add_users(testify_irc_channel_get(handle, channel), payload);
				break;
				case TESTIFY_IRC_COMMAND_QUIT :
				case TESTIFY_IRC_COMMAND_PART :
					channel_struct = testify_irc_channel_get(handle, channel);
					if(channel_struct != NULL)
						testify_irc_channel_remove_users(channel, name);
					if(command_func != NULL)
						command_func(op, name, channel, payload, user);		
				break;
				case TESTIFY_IRC_COMMAND_PRIVMSG :
				case TESTIFY_IRC_COMMAND_NOTICE : 
					channel_struct = testify_irc_channel_get(handle, channel);
					testify_irc_message_add(handle, channel_struct, name, payload);		
					if(command_func != NULL)
						command_func(op, name, channel, payload, user);			
				break;
				case TESTIFY_IRC_COMMAND_JOIN :
					channel_struct = testify_irc_channel_get(handle, channel);
					if(channel_struct == NULL)
					{
						testify_irc_channel_greate(handle, channel);
						sprintf(buffer, "NAMES [%s]\r\n", channel);
						for(i = 0; buffer[i] != 0; i++);
						testify_stream_send(handle, buffer, i);					 	 
					}else
						testify_irc_channel_add_users(channel_struct, name);
					if(command_func != NULL)
						command_func(op, name, channel, payload, user);
				break;
				default :
				//	printf("Command:%s ", command);
					if(command_func != NULL)
						command_func(op, name, channel, payload, user);
				break;
			}
			irc->buffer_used -= gap;
			for(i = 0; i < irc->buffer_used; i++)
				irc->buffer[i] = irc->buffer[i + gap];
		}else
			break;
	}
	return handle; 
}
