/*
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include "forge.h"
#include "imagine.h"
#define TESTIFY_INTERNAL
#include "testify.h"


#if defined _WIN32
#include <winsock2.h>
#include <Ws2tcpip.h>
#include <ws2bth.h>
#pragma comment(lib, "Ws2_32.lib")
#pragma comment(lib, "IPHLPAPI.lib")
//#include <stdio.h>
typedef unsigned int uint;
typedef SOCKET VSocket;
#else
typedef int VSocket;
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
#include <ifaddrs.h>
#include <net/if.h>
#define SOCKET_ERROR -1
#endif


#include "testify.h"

extern void testify_status_initialize();
extern void testify_status_portmap_set(TestifyHandle *handle);
extern void testify_status_portmap_remove(TestifyHandle *handle);

void testify_socket_destroy(uint32 socket)
{
#if defined _WIN32
	closesocket(socket);
#else
	close(socket);
#endif
}

#if defined _WIN32

boolean testify_socket_init_win32()
{
	static boolean initialized = FALSE;
	if(!initialized)
	{
		WSADATA wsaData;
		if(WSAStartup(MAKEWORD(2, 2), &wsaData) != 0)
		{
			fprintf(stderr, "Testify Error: WSAStartup failed.\n");
			return FALSE;
		}
		initialized = TRUE;
	}
	return TRUE;
}

#endif


void testify_address_print(TestifyAddress *address, char *buffer)
{
	switch(address->transport)
	{	
		case TESTIFY_TRANSPORT_NULL :
			sprintf(buffer, "NULL");
		break;
		case TESTIFY_TRANSPORT_IPV4 :
			sprintf(buffer, "%u.%u.%u.%u:%u", (address->address.ipv4 >> 24) & 0xFF, (address->address.ipv4 >> 16) & 0xFF, (address->address.ipv4 >> 8) & 0xFF, address->address.ipv4 & 0xFF,  address->port);
		break;
		case TESTIFY_TRANSPORT_IPV6 :
			sprintf(buffer, "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x:%u", address->address.ipv6[0], address->address.ipv6[1], address->address.ipv6[2], address->address.ipv6[3], address->address.ipv6[4], address->address.ipv6[5], address->address.ipv6[6], address->address.ipv6[7], address->port);
		break;
		case TESTIFY_TRANSPORT_BLUETOOTH :
			sprintf(buffer, "%02x:%02x:%02x.%02x:%02x:%02x", address->address.bluetooth[0], address->address.bluetooth[1], address->address.bluetooth[2], address->address.bluetooth[3], address->address.bluetooth[4], address->address.bluetooth[5]);
		break;
	}
}

boolean testify_address_compare_without_port(TestifyAddress *a, TestifyAddress *b)
{
	uint i;
	if(a->transport != b->transport)
		return FALSE;
	switch(a->transport)
	{
		case TESTIFY_TRANSPORT_NULL :
			return FALSE;
		case TESTIFY_TRANSPORT_IPV4 :
			return a->address.ipv4 == b->address.ipv4;
		case TESTIFY_TRANSPORT_IPV6 :
			for(i = 0; i < 8 && a->address.ipv6[i] == b->address.ipv6[i]; i++);
			return i == 8;
		case TESTIFY_TRANSPORT_BLUETOOTH :
			for(i = 0; i < 6 && a->address.bluetooth[i] == b->address.bluetooth[i]; i++);
			return i == 6;
	}
	return FALSE;
}

boolean testify_address_compare(TestifyAddress *a, TestifyAddress *b)
{
	char buffer[TESTIFY_ADDRESS_MAX_STRING_SIZE];
//	testify_address_print(a, buffer);
//	printf("Compare %u %s ", (uint)a->transport, buffer);
//	testify_address_print(b, buffer);
//	printf("- %u %s", (uint)b->transport, buffer);
	if(a->port != b->port && a->port != 0 && b->port != 0)
	{
//		printf("EARLY\n");
		return FALSE;
	}
	if(testify_address_compare_without_port(a, b))
	{
//		printf("TRUE\n");
		return TRUE;
	}else
	{
//		printf("FALSE\n");
		return FALSE;
	}
}



void testify_transport_translation(TestifyTransport transport, uint32 *output_transport)
{
	switch(transport)
	{
		case TESTIFY_TRANSPORT_IPV4 :
			*output_transport = AF_INET;
		break;
		case TESTIFY_TRANSPORT_IPV6 :
			*output_transport = AF_INET6;
		break;
#ifdef _WIN32
		case TESTIFY_TRANSPORT_BLUETOOTH :
			*output_transport = AF_BTH;
		break;
#endif
	}
}

void testify_protocol_translation(TestifyProtocol protocol, uint32 *output_protocol, uint32 *output_socket_type)
{
	switch(protocol)
	{
		case TESTIFY_PROTOCOL_DATAGRAM :
			*output_protocol = IPPROTO_UDP;
			*output_socket_type = SOCK_DGRAM;
		break;
		case TESTIFY_PROTOCOL_STREAM :
			*output_protocol = IPPROTO_TCP;
			*output_socket_type = SOCK_STREAM;
		break;
		case TESTIFY_PROTOCOL_RAW :
			*output_protocol = IPPROTO_RAW;
			*output_socket_type = SOCK_RAW;
		break;
	}
}

boolean testify_sockadrr_to_address(TestifyAddress *output_address, struct sockaddr_in *input)
{
	struct sockaddr_in *socket_address_ipv4;
	struct sockaddr_in6 *socket_address_ipv6;
#ifdef _WIN32
	SOCKADDR_BTH *socket_address_bluetooth;
#endif
	struct addrinfo hints, *result;
	uint i;
	if(input->sin_family == AF_INET)
	{
		socket_address_ipv4 = (struct sockaddr_in4 *)input;
		output_address->transport = TESTIFY_TRANSPORT_IPV4;
		output_address->port = ntohs(socket_address_ipv4->sin_port);
		output_address->address.ipv4 = ntohl(socket_address_ipv4->sin_addr.s_addr);
		return TRUE;
	}
	if(input->sin_family == AF_INET6)
	{
		socket_address_ipv6 = (struct sockaddr_in6 *)input;
		output_address->transport = TESTIFY_TRANSPORT_IPV6;
		output_address->port = ntohs(socket_address_ipv6->sin6_port);

		for(i = 0; i < 8; i++)
#ifdef _WIN32
			output_address->address.ipv6[i] = ntohs(socket_address_ipv6->sin6_addr.u.Word[i]);
#else
			output_address->address.ipv6[i] = ntohs((uint16)socket_address_ipv6->sin6_addr.s6_addr[i * 2] | ((uint16)socket_address_ipv6->sin6_addr.s6_addr[i * 2 + 1] << 8));
#endif
		return TRUE;
	}
#ifdef _WIN32
	if(input->sin_family == AF_BTH)
	{
		socket_address_bluetooth = (SOCKADDR_BTH *)input;
		output_address->transport = TESTIFY_TRANSPORT_BLUETOOTH;
		output_address->address.bluetooth[0] = ((uint64)socket_address_bluetooth->btAddr) & 0xFF;
		output_address->address.bluetooth[1] |= ((uint64)socket_address_bluetooth->btAddr >> 8) & 0xFF;
		output_address->address.bluetooth[2] |= ((uint64)socket_address_bluetooth->btAddr >> 16) & 0xFF;
		output_address->address.bluetooth[3] |= ((uint64)socket_address_bluetooth->btAddr >> 24) & 0xFF;
		output_address->address.bluetooth[4] |= ((uint64)socket_address_bluetooth->btAddr >> 32) & 0xFF;
		output_address->address.bluetooth[5] |= ((uint64)socket_address_bluetooth->btAddr >> 40) & 0xFF;
		return TRUE;
	}
#endif
	return FALSE;
}

size_t testify_address_to_sockadrr(TestifyAddress *address, uint8 *buffer)
{
	struct sockaddr_in *socket_address_ipv4;
	struct sockaddr_in6 *socket_address_ipv6;
#ifdef _WIN32
	SOCKADDR_BTH *socket_address_bluetooth;
#endif
	uint i;
#ifdef _WIN32
	if(address->transport == TESTIFY_TRANSPORT_BLUETOOTH)
	{
		socket_address_bluetooth = (SOCKADDR_BTH *)buffer;
		memset(socket_address_bluetooth, 0, sizeof *socket_address_bluetooth);
		socket_address_bluetooth->addressFamily = AF_BTH;
		socket_address_bluetooth->btAddr = (uint64)address->address.bluetooth[0] |
										((uint64)address->address.bluetooth[1]) << 8 |
										((uint64)address->address.bluetooth[2]) << 16 |
										((uint64)address->address.bluetooth[3]) << 24 |
										((uint64)address->address.bluetooth[4]) << 32 |
										((uint64)address->address.bluetooth[5]) << 40;
		socket_address_bluetooth->port = address->port;
		return sizeof *socket_address_bluetooth;
	}else
#endif
	if(address->transport == TESTIFY_TRANSPORT_IPV4)
	{
		socket_address_ipv4 = (struct sockaddr_in *)buffer;
		memset(socket_address_ipv4, 0, sizeof *socket_address_ipv4); /* Zero out structure */
		socket_address_ipv4->sin_family = AF_INET; /* host byte order */
		socket_address_ipv4->sin_port = htons(address->port); /* short, network byte order */
		if(address->address.ipv4 != 0)
			socket_address_ipv4->sin_addr.s_addr = htonl(address->address.ipv4);
		else
			socket_address_ipv4->sin_addr.s_addr = htonl(INADDR_ANY);
		return sizeof *socket_address_ipv4;
	}else
	{
		socket_address_ipv6 = (struct sockaddr_in6 *)buffer;
		memset(socket_address_ipv6, 0, sizeof *socket_address_ipv6); /* Zero out structure */
		socket_address_ipv6->sin6_family = AF_INET6; /* host byte order */
		socket_address_ipv6->sin6_port = htons(address->port); /* short, network byte order */
		for(i = 0; i < 8; i++)
#ifdef _WIN32
			socket_address_ipv6->sin6_addr.u.Word[i] = htons(address->address.ipv6[i]);
#else
		{
			uint16 tmp;
			tmp = htons(address->address.ipv6[i]);
			socket_address_ipv6->sin6_addr.s6_addr[i * 2] = (tmp >> 8) & 0xFF;
			socket_address_ipv6->sin6_addr.s6_addr[i * 2 + 1] = tmp & 0xFF;
		}
#endif
		return sizeof *socket_address_ipv6;
	}
}

uint32 testify_socket_create(TestifyAddress *address, TestifyProtocol protocol, uint32 flags)
{
	struct sockaddr_in6 socket_address_bind;
	struct sockaddr_in *socket_address_ipv4;
	struct sockaddr_in6 *socket_address_ipv6;
	socklen_t len;
#ifdef _WIN32
	SOCKADDR_BTH *socket_address_bluetooth;
#endif
	uint8 socket_address[1024];
	size_t socket_size;
    int buffer_size = 1 << 20;
	int option = TRUE;
	uint32 i, s, ipv6, prot, transport, socket_type, output;
#ifdef _WIN32
	if(!testify_socket_init_win32())
		return -1;
#endif
	testify_protocol_translation(protocol, &prot, &socket_type);
#ifdef _WIN32
	if(address->transport == TESTIFY_TRANSPORT_BLUETOOTH)
		prot = BTHPROTO_RFCOMM;
#endif
	testify_transport_translation(address->transport, &transport);
	// SOCK_NONBLOCK
	if((s = socket(transport, socket_type, prot)) == -1)
		return -1;

	if(transport == AF_INET6)	
	{
		if(setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, (char*)&ipv6, sizeof(ipv6)) == SOCKET_ERROR)
		{
			testify_socket_destroy(s);
			return -1;
		}
	}


	if(protocol == TESTIFY_PROTOCOL_DATAGRAM || protocol == TESTIFY_PROTOCOL_RAW)
		if(address->transport == TESTIFY_TRANSPORT_IPV4 || address->transport == TESTIFY_TRANSPORT_IPV6)
			if(setsockopt(s, SOL_SOCKET, SO_BROADCAST, &option, sizeof option) != 0)
				fprintf(stderr, "Testify: Couldn't set brodcast option of socket to %d\n", buffer_size);
	if(setsockopt(s, SOL_SOCKET, SO_SNDBUF, &buffer_size, sizeof buffer_size) != 0)
		fprintf(stderr, "Testify: Couldn't set send buffer size of socket to %d\n", buffer_size);
	if(setsockopt(s, SOL_SOCKET, SO_RCVBUF, &buffer_size, sizeof buffer_size) != 0)
		fprintf(stderr, "Testify: Couldn't set receive buffer size of socket to %d\n", buffer_size);


	socket_size = testify_address_to_sockadrr(address, socket_address);
	if(flags & TESTIFY_FLAG_BIND)
	{
		if(bind(s, (struct sockaddr *)socket_address, socket_size) != 0)
		{
#ifdef _WIN32
			printf("Testify: bind failed with error %u port number %u\n", WSAGetLastError(), (uint)address->port);
#elif defined( __linux__ )
			printf("Testify: bind failed with error %u port number %u\n", errno, (uint)address->port);
#endif
			testify_socket_destroy( s );
			return -1;
		}//else
		//	printf("Testify: successfuly bound port %u\n", (uint)address->port);

		len = sizeof(socket_address_bind);
		if(getsockname(s, (struct sockaddr *)&socket_address_bind, &len) != -1)
		{
			if(!testify_sockadrr_to_address(address, (struct sockaddr *)&socket_address_bind))
				len = 0;
		}
	}else if(prot != IPPROTO_UDP)
	{	

	#if defined _WIN32
		if(flags & TESTIFY_FLAG_NON_BLOCKING_CONNECT)
		{
			unsigned long	one = 1UL;
			if(ioctlsocket(s, FIONBIO, &one) != 0)
				return -1;
		}
	#else
		if(fcntl(s, F_SETFL, O_NONBLOCK) != 0) /* always set non-blocking on non windows systems to limit long timeouts */
		{
			fprintf(stderr, "Testify: Couldn't make socket non-blocking\n");
			exit(0); /* FIX ME */
			return -1;
		}
	#endif
		if(connect(s, (struct sockaddr *)&socket_address, socket_size) < 0)
		{
#ifdef _WIN32
			printf("Testify: connect failed with error %u\n", WSAGetLastError());
			testify_socket_destroy(s);
			return -1;
#else 

			struct timeval	tv;
			unsigned int time;
			fd_set fd_write;
			FD_ZERO(&fd_write);
			FD_SET(s, &fd_write);
			tv.tv_sec  = 5;
			tv.tv_usec = 0;
			time = select(s + 1, NULL, &fd_write, NULL, &tv);	
			if(!FD_ISSET(s, &fd_write))
			{
				printf("Testify: connect failed with error %u time = %u\n", errno, time);
				testify_socket_destroy(s);
				return -1;		
			}
#endif
		}
		if(TESTIFY_AT_GLOBAL == testify_address_type(address)) 
			testify_status_flags_set(TESTIFY_NS_INTERNET_ACCESS_CONFIRMED, 0);
	}
	{
#if defined _WIN32
		unsigned long one;
		if((~flags) & TESTIFY_FLAG_BLOCKING)
			one = 1;
		else
			one = 0;
		if(ioctlsocket(s, FIONBIO, &one) != 0)
		{
		//	fprintf(stderr, "Testify: Couldn't make socket non-blocking\n");
			testify_socket_destroy(s);
			return -1;
		}
#else		
		if((~flags) & TESTIFY_FLAG_BLOCKING)
		{
			if(fcntl(s, F_SETFL, O_NONBLOCK) != 0)
			{
				fprintf(stderr, "Testify: Couldn't make socket non-blocking\n");
				exit(0); /* FIX ME */
				return -1;
			}
		}
#endif
	}

	return s;
}

TestifyHandle *testify_create(TestifyAddress *address, TestifyProtocol protocol, uint64 flags)
{
	TestifyHandle *output;
	int socket;
	uint i;
	testify_status_initialize();
	socket = testify_socket_create(address, protocol, flags);
	if(socket == -1)
		return NULL;
	output = malloc(sizeof *output);
	output->socket = socket;
	output->bound_to_specific_ip = TRUE;
	if(address == NULL || address->address.ipv4 == 0)
		output->bound_to_specific_ip = FALSE;
	output->address = *address;
	output->mapping_count = 0;
	output->connected = TRUE;
	output->protocol = protocol;
	output->custom = NULL;
	if(address->transport != TESTIFY_TRANSPORT_BLUETOOTH)
	{
		if(flags & TESTIFY_FLAG_PORT_MAPPED)
		{
			testify_status_portmap_set(output);
		}
	}
	return output;
}


void testify_destroy(TestifyHandle *handle)
{
	testify_socket_destroy(handle->socket);
//	if(flags & TESTIFY_FLAG_PORT_MAPPED)
	testify_status_portmap_remove(handle);
	if(handle->custom != NULL)
		testify_irc_destroy(handle->custom);
	free(handle);
}

// TestifyTransport transport

boolean testify_address_lookup(TestifyAddress *address, char *dns_name, TestifyTransport transport, uint16 default_port)
{
	struct sockaddr_in *socket_address_ipv4;
	struct sockaddr_in6 *socket_address_ipv6;
	struct addrinfo hints, *result;
	char *colon = NULL, *buf = NULL, *loopback = "loopback", *broadcast = "broadcast", *localhost = "localhost", *any = "any";
	uint i, j;
	uint32 prot, socket_type, socket_transport;
	char buffer[1024];
	address->transport = transport;
	address->port = default_port;
	if(transport == TESTIFY_TRANSPORT_BLUETOOTH)
	{
		for(i =  0; localhost[i] != 0 && dns_name[i] == localhost[i]; i++);
		if(dns_name[i] == localhost[i])
		{
			for(j = 0; j < 6; j++)
				address->address.bluetooth[j] = 0;
		}else
		{
			for(i = j = 0; TRUE; i++)
			{
				if(dns_name[i] >= '0' && dns_name[i] <= '9')
					address->address.bluetooth[j] = (dns_name[i] - '0') << 4;
				else if(dns_name[i] >= 'A' && dns_name[i] <= 'F')
					address->address.bluetooth[j] = (dns_name[i] + 10 - 'A') << 4;
				else if(dns_name[i] >= 'a' && dns_name[i] <= 'f')
					address->address.bluetooth[j] = (dns_name[i] + 10 - 'a') << 4;
				else
					return FALSE;
				i++;
				if(dns_name[i] >= '0' && dns_name[i] <= '9')
					address->address.bluetooth[j] |= dns_name[i] - '0';
				else if(dns_name[i] >= 'A' && dns_name[i] <= 'F')
					address->address.bluetooth[j] |= dns_name[i] + 10 - 'A';
				else if(dns_name[i] >= 'a' && dns_name[i] <= 'f')
					address->address.bluetooth[j] |= dns_name[i] + 10 - 'a';
				else
					return FALSE;
				i++;
				if(i < 6 * 3 - 1)
				{
					if(dns_name[i] != ':')
						return -1;
					return TRUE;
				}else
					break;
				j++;
			}
		}
		if(dns_name[i++] == ':')
			for(address->port = 0; dns_name[i] >= '0' && dns_name[i] <= '9'; i++)
				address->port = (address->port * 10) + (dns_name[i] - '0');
		return TRUE;
	}else
	{
#if defined _WIN32
		if(!testify_socket_init_win32())
			return FALSE;
#endif
		memset(&hints, 0, sizeof hints);
		memset(&result, 0, sizeof result);
		if(dns_name != NULL)
		{
			for(i = 0; dns_name[i] != 0 && i < 1023; i++)
			{
				if(dns_name[i] == ':')
				{
					memcpy(buffer, dns_name, i);
					buffer[i] = 0;
					default_port = 0;
					for(i++; dns_name[i] >= '0' && dns_name[i] <= '9'; i++)
						default_port = (default_port * 10) + (dns_name[i] - '0');
					dns_name = buffer;
					break;
				}
			}


			for(i = 0; dns_name[i] == any[i] && any[i] != 0; i++);
			if(any[i] == 0)
			{
				address->transport = TESTIFY_TRANSPORT_IPV4;
				address->address.ipv4 = 0;
				return TRUE;
			}
			for(i = 0; dns_name[i] == broadcast[i] && broadcast[i] != 0; i++);
			if(broadcast[i] == 0)
			{
				address->transport = TESTIFY_TRANSPORT_IPV4;
				address->address.ipv4 = ((uint32)224 << (uint32)24) + (uint32)1; // 224.0.0.1 broadcast address;
				return TRUE;
			}
			for(i = 0; dns_name[i] == loopback[i] && loopback[i] != 0; i++);
			if(loopback[i] == 0)
				dns_name = NULL;
		}else
		{
			address->transport = TESTIFY_TRANSPORT_IPV4;
			address->address.ipv4 = 0;
			return TRUE;
		}




	//	testify_protocol_translation(address->protocol, &transport, &prot, &socket_type);
		//	*output_transport = AF_INET;
		//	*output_protocol = IPPROTO_UDP;

		testify_transport_translation(transport, &socket_transport);
		hints.ai_family = socket_transport;
		hints.ai_socktype = 0;
		hints.ai_protocol = 0;
		hints.ai_flags = AI_CANONNAME;
		if(0 == getaddrinfo(dns_name, NULL, &hints, &result))
		{
			testify_sockadrr_to_address(address, result->ai_addr);
			address->port = default_port;

		/*	if(result->ai_family == AF_INET)
			{
				socket_address_ipv4 = (struct sockaddr_in4 *)result->ai_addr;
				address->address.ipv4 = ntohl(socket_address_ipv4->sin_addr.s_addr);
				return TRUE;
			}
			if(result->ai_family == AF_INET6)
			{
				socket_address_ipv6 = (struct sockaddr_in6 *)result->ai_addr;
				for(i = 0; i < 8; i++)
					address->address.ipv6[i] = ntohs(socket_address_ipv6->sin6_addr.u.Word[i]);
				return TRUE;
			}*/
			return TRUE;
		}
#ifdef _WIN32
		printf("Testify: getaddrinfo failed with error %u\n", WSAGetLastError());
#endif
		return FALSE;
	}
}


TestifyHandle *testify_wait_for_connection(TestifyHandle *listener)
{
	TestifyHandle *handle;
	TestifyAddress from;
	struct sockaddr_in address;
	int socket;
	uint length;
	boolean read = TRUE;
	if(listen(listener->socket, 16) >= 0)
	{
		//testify_network_wait(&listener, &read, NULL, 1, 0);
		if(read)
		{
			length = sizeof(struct sockaddr_in);
			if((socket = accept(listener->socket, (struct sockaddr *) &address, &length)) >= 0)
			{
				printf("trying to recive connection\n");
#if defined _WIN32
				{
					unsigned long	one = 1UL;
					if(ioctlsocket(socket, FIONBIO, &one) != 0)
						return NULL;
				}
#else
				if(fcntl(socket, F_SETFL, O_NONBLOCK) != 0)
				{
					fprintf(stderr, "Testify: Couldn't make socket non-blocking\n");
					exit(0); /* FIX ME */
					return NULL;
				}
#endif
				if(testify_sockadrr_to_address(&from, &address))
				{
					if(TESTIFY_AT_GLOBAL == testify_address_type(&address))
						testify_status_flags_set(TESTIFY_NS_INTERNET_ACCESS_CONFIRMED, 0);
					from.transport = from.transport;
					from.port = ntohs(address.sin_port);
					handle = malloc(sizeof *handle);
					handle->custom = NULL;
					handle->address = from;
					handle->socket = socket;
					handle->connected = TRUE;
					handle->flags = 0;
					handle->protocol = TESTIFY_PROTOCOL_STREAM;
					handle->mapping_count = 0;
					handle->bound_to_specific_ip = 0;
					handle->next_mapped_handle = NULL;
					handle->custom = NULL;
					return handle;
				}
			}
		}
	}//else
	//	listener->connected = FALSE;
	return NULL;
}

boolean	testify_connected(TestifyHandle *handle)
{
	return handle->connected;
}

size_t testify_stream_recive(TestifyHandle *handle, uint8 *buffer, size_t buffer_size)
{
	int output;
	output = recv(handle->socket, buffer, buffer_size, 0);
	if(output < 0)
	{
		int error;
#if defined _WIN32
		error = WSAGetLastError();
		if(error == WSAEWOULDBLOCK)
			return 0;
#else
		error = errno;
		if(error ==  EAGAIN || error == EWOULDBLOCK || error == 0) /* Given that its non blocking recv may return an error if no data is available. EWOULDBLOCK */
			return 0;
	printf("Testify: testify_stream_recive Connection error %i ", error);
#endif
		handle->connected = FALSE;
		output = 0;
	}
	if(output != 0)
		output += 0;
	return (size_t)output;
}

size_t testify_stream_send(TestifyHandle *handle, uint8 *buffer, size_t buffer_size)
{
	int size;
	size = send(handle->socket, buffer, buffer_size, 0);
	if(size < 0)
	{
		int error;
#if defined _WIN32
		error = WSAGetLastError();
		if(error == WSAEWOULDBLOCK)
			return 0;
#else
		error = errno;
		if(error ==  EAGAIN || error == EWOULDBLOCK || error == 0) /* Given that its non blocking recv may return an error if no data is available. EWOULDBLOCK */
			return 0;
	printf("Testify: testify_stream_send Connection error %i ", error);
#endif
		handle->connected = FALSE;
		return 0;
	}
	return (size_t)size;
}

size_t testify_datagram_send(TestifyHandle *handle, TestifyAddress *to, uint8 *buffer, size_t size)
{
	struct sockaddr_in	address_in;
	uint8 socket_address[1024];
	TestifyAddress broadcast;
	TestifyServiceState status;
	size_t size_of;
	int i, out;

	if(to->address.ipv4 == ((uint32)224 << (uint32)24) + (uint32)1 && to->transport == TESTIFY_TRANSPORT_IPV4)
	{
		out = 0;
		for(i = 0; testify_status_service_get(i, TESTIFY_SST_BROADCAST, NULL, 0, &broadcast, &status); i++)
		{
			if(status != TESTIFY_SS_FAILED)
			{
				broadcast.port = to->port;
				size_of = testify_address_to_sockadrr(&broadcast, socket_address);
				out = sendto(handle->socket, buffer, size, 0, (struct sockaddr *)&socket_address, size_of);
			}
		}
		return out;
	}
	size_of = testify_address_to_sockadrr(to, socket_address);
	out = sendto(handle->socket, buffer, size, 0, (struct sockaddr *)&socket_address, size_of);
	if(out < 0)
	{
#ifdef _WIN32
		printf("Testify: sendto failed with error %u\n", WSAGetLastError());
#endif
		handle->connected = FALSE;
		return 0;
	}
	return out;
}

#define testify_ipv4_address(a, b, c, d) (a * (256 * 256 * 256) + b * (256 * 256) + c * 256 + d)


size_t testify_datagram_receive(TestifyHandle *handle, TestifyAddress *from, uint8 *buffer, size_t size)
{
	struct sockaddr_in address_in;
	int from_length = sizeof address_in, len;
	memset(&address_in, 0, sizeof address_in);
	address_in.sin_family = AF_INET;
	address_in.sin_port = htons(handle->address.port);
	address_in.sin_addr.s_addr = INADDR_ANY;
	len = recvfrom(handle->socket, buffer, size, 0, (struct sockaddr *) &address_in, &from_length);
	if(len < 0)
	{
#ifdef _WIN32
		len = WSAGetLastError();
		if(10035 != len) // Non blocking 
			len = 0;
#endif
		return 0;
	}
	if(testify_sockadrr_to_address(from, &address_in))
		return (size_t)len;
	return 0;
}



uint testify_wait(TestifyHandle **handles, boolean *read, boolean *write, unsigned int handle_count, unsigned int microseconds)
{
	struct timeval	tv;
	fd_set fd_read, fd_write;
	unsigned int i, max = 0;

	FD_ZERO(&fd_read);
	FD_ZERO(&fd_write);
	for(i = 0; i < handle_count; i++)
	{
		if(read != NULL && read[i] && handles[i]->connected)
		{
			FD_SET(handles[i]->socket, &fd_read);
			if(max < handles[i]->socket)
				max = handles[i]->socket;
		}
		if(write != NULL && write[i] && handles[i]->connected)
		{
			FD_SET(handles[i]->socket, &fd_write);
			if(max < handles[i]->socket)
				max = handles[i]->socket;
		}
	}
	tv.tv_sec  = microseconds / 1000000;
	tv.tv_usec = microseconds % 1000000;
	select(max + 1, &fd_read, &fd_write, NULL, &tv);
	for(i = 0; i < handle_count; i++)
	{
		if(read != NULL && read[i] && handles[i]->connected)
			read[i] = FD_ISSET(handles[i]->socket, &fd_read);
		if(write != NULL && write[i] && handles[i]->connected)
			write[i] = FD_ISSET(handles[i]->socket, &fd_write);
	}
	return microseconds;
}

void testify_pack_swap16(uint8 *output, uint8 *input)
{
	output[0] = input[1];
	output[1] = input[0];
}

void testify_pack_swap32(uint8 *output, uint8 *input)
{
	output[0] = input[3];
	output[1] = input[2];
	output[2] = input[1];
	output[3] = input[0];
}

void testify_pack_swap64(uint8 *output, uint8 *input)
{
	output[0] = input[7];
	output[1] = input[6];
	output[2] = input[5];
	output[3] = input[4];
	output[4] = input[3];
	output[5] = input[2];
	output[6] = input[1];
	output[7] = input[0];
}

extern void testify_pack_swap16(uint8 *output, uint8 *input);
extern void testify_pack_swap32(uint8 *output, uint8 *input);
extern void testify_pack_swap64(uint8 *output, uint8 *input);

size_t testify_address_serialize(TestifyAddress *address, uint8 *buffer)
{
	uint i;
	buffer[0] = address->transport;
	testify_pack_littleendian16(&buffer[1], &address->port);
	switch(address->transport)
	{
		case TESTIFY_TRANSPORT_NULL :
		return 3;
		case TESTIFY_TRANSPORT_IPV4 :
			testify_pack_littleendian32(&buffer[3], &address->address.ipv4);
		return 7;
		case TESTIFY_TRANSPORT_IPV6 :
			for(i = 0; i < 8; i++)
				testify_pack_littleendian16(&buffer[3 + 2 * i], &address->address.ipv6[i]);
		return 3 + 8 * 2;
		case TESTIFY_TRANSPORT_BLUETOOTH :
			memcpy(&buffer[3], address->address.bluetooth, 6);
		return 3 + 6;
		default :
			printf("TESTIFY: Error: Trying to serialize address with unknown transport.");
		return 0;
	}
	return 0;
}
size_t testify_address_deserialize(TestifyAddress *address, uint8 *buffer, uint buffer_length)
{
	uint i;
	if(buffer_length < 3)
		return 0;
	address->transport = buffer[0] & 7;
	testify_pack_littleendian16(&address->port, &buffer[1]);
	switch(address->transport)
	{
		case TESTIFY_TRANSPORT_NULL :
		return 3;
		case TESTIFY_TRANSPORT_IPV4 :
			if(buffer_length < 3 + 4)
				return 0;
			testify_pack_littleendian32(&address->address.ipv4, &buffer[3]);
		return 7;
		case TESTIFY_TRANSPORT_IPV6 :
			if(buffer_length < 3 + 8 * 2)
				return 0;
			for(i = 0; i < 8; i++)
				testify_pack_littleendian16(&address->address.ipv6[i], &buffer[3 + 2 * i]);
		return 3 + 8 * 2;
		case TESTIFY_TRANSPORT_BLUETOOTH :
			if(buffer_length < 3 + 6)
				return 0;
			memcpy(address->address.bluetooth, &buffer[3], 6);
		return 3 + 6;
		default :
			printf("TESTIFY: Error: Trying to deserialize address with unknown transport.");
		return 0;
	}
}
