/*
Test if a local address could be Static/Public
if(maybe public)
	DNS check.
if(Not static)
{
	-Try PCP
	if(IP address)
		Confirm IP with DNS
}

When connecting:
	look at internal lists of dosovery.

if()

{
	-If(available)  
	{
		-Take IP port.
		-Test DNS
	}
}

-IP
-Hole punch required.
*
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include "forge.h"
#include "imagine.h"
#define TESTIFY_INTERNAL
#include "testify.h"

typedef enum{
	T_NATPMPRT_SUCSESS = 0,
	T_NATPMPRT_UNSUPORTED_VERSION = 1,
	T_NATPMPRT_REFUSED = 2,
	T_NATPMPRT_NETWORK_FAILURE = 3,
	T_NATPMPRT_NO_REASORCES = 4,
	T_NATPMPRT_UNSUPPORTED_OP = 5,
}TestifyNATPMPReturnCodes;

typedef enum{
	T_PCP_SUCSESS = 0,
	T_PCP_UNSUPORTED_VERSION = 1,
	T_PCP_NOT_AUTHORIZED = 2,
	T_PCP_MALFORMED_REQUEST = 3,
	T_PCP_UNSUPP_OPCODE = 4,
	T_PCP_UNSUPP_OPTION = 5,
	T_PCP_MALFORMED_OPTION = 6,
	T_PCP_NETWORK_FAILURE = 7,
	T_PCP_NO_RESOURCES = 8,
	T_PCP_UNSUPP_PROTOCOL = 9,
	T_PCP_USER_EX_QUOTA = 10,
	T_PCP_CANNOT_PROVIDE_EXTERN = 11,
	T_PCP_ALADDRESS_MISMATCH = 12,
	T_PCP_EXCESSIVE_REMOTE_PEERS = 13
}TestifyPCPReturnCodes;

#define testify_ipv4_address(a, b, c, d) (a * (256 * 256 * 256) + b * (256 * 256) + c * 256 + d)

void testify_portmap_send(TestifyHandle *handle, TestifyAddress *gateway, TestifyProtocol protocol, uint version, uint timer, TestifyAddress *forward)
{
	uint i, j, k, size, time, adapter_count;
	uint8 buffer[1500];
	uint32 tmp32;
	TestifyAddress to;
	to = *gateway;
	to.port = 5351;
	switch(version)
	{
		case 0 :
			buffer[0] = version; /* version */
			buffer[1] = 0; /* Op code : request external IP address */
			testify_datagram_send(handle, &to, buffer, 2);
			buffer[0] = version; /* version */
			if(protocol == TESTIFY_PROTOCOL_DATAGRAM) 
				buffer[1] = 1; /* Op code : protocol UDP */
			else
				buffer[1] = 2; /* Op code : protocol TCP */
			buffer[2] = 0; /* Reserved */
			buffer[3] = 0; /* Reserved */
			testify_pack_bigendian16(&buffer[4], &forward->port); /* Internal Port */
			testify_pack_bigendian16(&buffer[6], &forward->port); /* External Port */
			tmp32 = (uint)7200;
			testify_pack_bigendian32(&buffer[8], &tmp32);
			testify_datagram_send(handle, &to, buffer, 12);
		break;
		case 2 :
			buffer[0] = version; /* version */
			buffer[1] = 1; /* Op code : map (there is no way to get external IP from PCP )*/
			buffer[2] = 0; /* Reserved */
			buffer[3] = 0; /* Reserved */
			tmp32 = 60;
			testify_pack_bigendian32(&buffer[4], &tmp32);
			if(forward->transport == TESTIFY_TRANSPORT_IPV4)
			{
				for(k = 0; k < 10; k++)
					buffer[8 + k] = 0; /* "IP" */
				buffer[18] = 255; /* "IP" */
				buffer[19] = 255; /* "IP" */							
				testify_pack_bigendian32(&buffer[20], &forward->address.ipv4);
			}else /* IP V6 */
				for(k = 0; k < 8; k++)
					testify_pack_bigendian16(&buffer[8 + 2 * k], &forward->address.ipv6[k]);
			/* Header end */
			for(k = 0; k < 12; k++)
				buffer[24 + k] = 0;
								
			if(TESTIFY_PROTOCOL_DATAGRAM == protocol) 
				buffer[36] = 17; /*UDP*/
			else
				buffer[36] = 6; /*TCP*/
			buffer[37] = 0; /* Reserved */
			buffer[38] = 0; /* Reserved */
			buffer[39] = 0; /* Reserved */
			testify_pack_bigendian16(&buffer[40], &forward->port); /* Internal Port */
			testify_pack_bigendian16(&buffer[42], &forward->port); /* External Port */
			if(forward->transport == TESTIFY_TRANSPORT_IPV4)
			{
				for(k = 0; k < 10; k++)
					buffer[44 + k] = 0; /* "IP" */
				buffer[54] = 255; /* "IP" */
				buffer[55] = 255; /* "IP" */	
				testify_pack_bigendian32(&buffer[20], &forward->address.ipv4);
			}else /* IP V6 */
				for(k = 0; k < 8; k++)
					testify_pack_bigendian16(&buffer[44 + 2 * k], &forward->address.ipv6[k]);
			testify_datagram_send(handle, &to, buffer, 60);
		break;
	}
}

typedef enum{
	TESTIFY_PMRRV_CHANGE_PROTOCOL_TO_NAT_PMP = 0,
	TESTIFY_PMRRV_CHANGE_PROTOCOL_TO_PCP = 2,
	TESTIFY_PMRRV_ERROR,
	TESTIFY_PMRRV_UNKNOWN_OP,
	TESTIFY_PMRRV_PORT_UPDATE,
	TESTIFY_PMRRV_IP_UPDATE,
	TESTIFY_PMRRV_MAPPING_UPDATE
}TestifyPortMapResponceReturnValue;

TestifyPortMapResponceReturnValue testify_portmap_buffer(uint8 *buffer, uint size, uint *timeout, uint16 *private_port, TestifyAddress *address)
{
	uint version, opcode;
	uint16 result, port;
	version = buffer[0]; /* version" */
	opcode = buffer[1]; /* "opcode" */
	testify_pack_bigendian16(&result, &buffer[2]);/* "result code" */
	switch(version)
	{
		case 0 : /* NAT PMP */
			if(result == T_NATPMPRT_UNSUPORTED_VERSION ||
				result == T_NATPMPRT_UNSUPPORTED_OP)
				return TESTIFY_PMRRV_CHANGE_PROTOCOL_TO_PCP; /* Try a differenet protocol */
			else if(result == T_NATPMPRT_SUCSESS && size >= 16)
			{

				if(opcode == 128 && size >= 16)
				{
				/*	testify_pack_bigendian32(&tmp, &buffer[4]); /* time */ 
					testify_pack_bigendian32(&address->address.ipv4, &buffer[8]);
					address->transport = TESTIFY_TRANSPORT_IPV4;
					
					testify_status.status_bits |= TESTIFY_NS_PORTMAPED_PMP + TESTIFY_NS_PORT_FORWARDING_AVAILABLE;

					return TESTIFY_PMRRV_IP_UPDATE;
				}
				else if(opcode == 128 || opcode == 129)
				{
				/*	testify_pack_bigendian32(timeout, &buffer[4]); /* time */ 
					testify_status.status_bits |= TESTIFY_NS_PORTMAPED_PMP + TESTIFY_NS_PORT_FORWARDING_AVAILABLE;
					testify_pack_bigendian16(&private_port, &buffer[8]); /* private port */ 		
					testify_pack_bigendian16(&address->port, &buffer[10]); /* public port */
					testify_pack_bigendian32(timeout, &buffer[12]); /* time */ 
					address->transport = TESTIFY_TRANSPORT_IPV4;
					return TESTIFY_PMRRV_PORT_UPDATE;
				}
				return TESTIFY_PMRRV_UNKNOWN_OP;
			}
			return TESTIFY_PMRRV_ERROR;
		break;
		case 2 : /* PCP */
			if(result == T_PCP_UNSUPORTED_VERSION ||
				result == T_PCP_MALFORMED_REQUEST ||
				result == T_PCP_UNSUPP_OPCODE ||
				result == T_PCP_UNSUPP_OPTION ||
				result == T_PCP_MALFORMED_OPTION ||
				result == T_PCP_UNSUPP_PROTOCOL)
				return TESTIFY_PMRRV_CHANGE_PROTOCOL_TO_NAT_PMP; /* Try a differenet protocol */
			else if(result == T_PCP_SUCSESS && size >= 40)
			{
				testify_pack_bigendian32(timeout, &buffer[4]); /* time */ 
				buffer[36]; /* tcp / udp */
				testify_pack_bigendian16(private_port, &buffer[40]); /* private port */ 
				if(opcode == 128 || opcode == 129)
				{ 
					testify_pack_bigendian16(&address->port, &buffer[42]); /* public port */	
					if(buffer[44] == 0 &&
						buffer[45] == 0 &&
						buffer[46] == 0 &&
						buffer[47] == 0 &&
						buffer[48] == 0 &&
						buffer[49] == 0 &&
						buffer[50] == 0 &&
						buffer[51] == 0 &&
						buffer[52] == 0 &&
						buffer[53] == 0 &&
						buffer[54] == 255 &&
						buffer[55] == 255)
					{
						address->transport = TESTIFY_TRANSPORT_IPV4;
						testify_pack_bigendian32(&address->address.ipv4, &buffer[56]);
					} else
					{
						address->transport = TESTIFY_TRANSPORT_IPV6;
						testify_pack_bigendian16(&address->address.ipv6[0], &buffer[44]);
						testify_pack_bigendian16(&address->address.ipv6[1], &buffer[46]);
						testify_pack_bigendian16(&address->address.ipv6[2], &buffer[48]);
						testify_pack_bigendian16(&address->address.ipv6[3], &buffer[50]);
						testify_pack_bigendian16(&address->address.ipv6[4], &buffer[52]);
						testify_pack_bigendian16(&address->address.ipv6[5], &buffer[54]);
						testify_pack_bigendian16(&address->address.ipv6[6], &buffer[56]);
						testify_pack_bigendian16(&address->address.ipv6[7], &buffer[58]);
					}
					
					testify_status.status_bits |= TESTIFY_NS_PORTMAPED_PCP + TESTIFY_NS_PORT_FORWARDING_AVAILABLE;
					return TESTIFY_PMRRV_MAPPING_UPDATE;
				}else
					return TESTIFY_PMRRV_UNKNOWN_OP;
			}			
			return TESTIFY_PMRRV_ERROR;
		break;
	}
	return TESTIFY_PMRRV_ERROR;
}

void testify_portmap_recive(TestifyStatus *status, TestifyAddress *from, uint8 *buffer, uint size)
{
	TestifyHandle *handle;
	uint16 private_port;
	TestifyAddress address;
	uint i, service, timeout, mapping = ~0;
	uint16 port;
	TestifyPortMapResponceReturnValue output;
	for(service = 0; status->services[service].type != TESTIFY_SST_GATEWAY || !testify_address_compare(from, &status->services[service].address);)
		if(++service == status->service_count)
			return;

	output = testify_portmap_buffer(buffer, size, &timeout, &private_port, &address);
	
	for(handle = status->mapped_ports; handle != NULL; handle = handle->next_mapped_handle)
		if(handle->address.port == private_port)
			break;
	if(handle != NULL)
		for(mapping = 0; mapping < handle->mapping_count && !testify_address_compare_without_port(&handle->mappings[mapping].gateway, from); mapping++);
	
	switch(output)
	{
		case TESTIFY_PMRRV_CHANGE_PROTOCOL_TO_NAT_PMP :			
		break;
		case TESTIFY_PMRRV_CHANGE_PROTOCOL_TO_PCP :
		break;
		case TESTIFY_PMRRV_ERROR :
			status->services[service].status = TESTIFY_SS_FAILED;
		break;
		case TESTIFY_PMRRV_UNKNOWN_OP :
			status->services[service].status = TESTIFY_SS_FAILED;
		break;
		case TESTIFY_PMRRV_PORT_UPDATE :
			if(handle != NULL && mapping < handle->mapping_count)
			{
				handle->mappings[mapping].address.port = address.port;
				handle->mappings[mapping].timeout = timeout - 5;
			}
				
		break;
		case TESTIFY_PMRRV_IP_UPDATE :		
			for(handle = status->mapped_ports; handle != NULL; handle = handle->next_mapped_handle)
			{
				for(mapping = 0; mapping < handle->mapping_count; mapping++)
				{
					if(testify_address_compare_without_port(&handle->mappings[mapping].gateway, from))
					{
						port = handle->mappings[mapping].address.port;
						handle->mappings[mapping].address = address;
						handle->mappings[mapping].address.port = port;
					}
				}
			}
		break;
		case TESTIFY_PMRRV_MAPPING_UPDATE :
			if(handle != NULL && mapping < handle->mapping_count)
			{
				handle->mappings[mapping].address = address;	
				handle->mappings[mapping].timeout = timeout - 5;
			}
		break;
	}

}

void testify_portmap_update(TestifyStatus *status)
{
	TestifyHandle *handle;
	TestifyAddress address;
	uint i, mapping, count;
	boolean adapter_on = FALSE, used[TESTIFY_HANDLE_ADDRESS_COUNT_MAX];
	for(handle = status->mapped_ports; handle != NULL; handle = handle->next_mapped_handle)
	{
		for(i = 0; i < handle->mapping_count; i++)
			used[i] = FALSE;
		for(i = 0; i < status->service_count; i++)
		{
			if(status->services[i].type == TESTIFY_SST_ADAPTER)
			{
				address = status->services[i].address;
				adapter_on = TRUE;
			}
			if(adapter_on && status->services[i].type == TESTIFY_SST_GATEWAY && status->services[i].status != TESTIFY_SS_FAILED)
			{
				address.port = handle->address.port;
				if(!handle->bound_to_specific_ip || testify_address_compare(&handle->address, &address))
				{
					for(mapping = 0; mapping < handle->mapping_count && !testify_address_compare_without_port(&handle->mappings[mapping].gateway, &status->services[i].address); mapping++);
					if(mapping < TESTIFY_HANDLE_ADDRESS_COUNT_MAX)
					{
						if(mapping == handle->mapping_count)
						{
							handle->mappings[mapping].address.transport = TESTIFY_TRANSPORT_NULL;
							handle->mappings[mapping].timeout = 0;
							handle->mappings[mapping].gateway = status->services[i].address;
							handle->mapping_count++;
						}
						used[mapping] = TRUE;
						if(handle->mappings[mapping].timeout == 0)
						{
							status->services[i].last_updated = 0;
							if(status->services[i].status == TESTIFY_SS_WAITING || status->services[i].status == TESTIFY_SS_SUCSESS)
								testify_portmap_send(status->handle, &status->services[i].address, handle->protocol, 2, 7200, &address);
							else
								testify_portmap_send(status->handle, &status->services[i].address, handle->protocol, 0, 7200, &address);
							if(status->services[i].status == TESTIFY_SS_WAITING && status->services[i].last_updated > 5.0)
								status->services[i].status = TESTIFY_SS_FAILED;
						}else
							handle->mappings[mapping].timeout--;
					}
				}
				adapter_on = FALSE;
			}
		}
		for(i = 0; i < handle->mapping_count; i++)
		{
			if(!used[i])
			{	
				used[i] = used[--handle->mapping_count];
				handle->mappings[i--] = handle->mappings[handle->mapping_count];
			}
		}
	//	status->mapped_ports[i].timer -= delta;
	}
}