/*
Test if a local address could be Static/Public
if(maybe public)
	DNS check.
if(Not static)
{
	-Try PCP
	if(IP address)
		Confirm IP with DNS
}

When connecting:
	look at internal lists of dosovery.

if()

{
	-If(available)
	{
		-Take IP port.
		-Test DNS
	}
}

-IP
-Hole punch required.
* 
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include "forge.h"
#include "imagine.h"
#define TESTIFY_INTERNAL
#include "testify.h"

boolean testify_status_init = FALSE;
TestifyStatus testify_status;


char *testify_network_status_strings[] = {"TESTIFY_NS_NOT_ONLINE, // We have no idea what kind of network we are on.",
										"TESTIFY_NS_UNKNOWN, // We have no idea what kind of network we are on.",
										"TESTIFY_NS_NAT, // We are behind a NAT network without port forwarding.",
										"TESTIFY_NS_NAT_UNSTABLE, // We are behind a NAT network without port forwarding and our IP keeps changing.",
										"TESTIFY_NS_NAT_PORTMAPED_UNSTABLE, // We are behind a NAT network with port forwarding but our IP keeps changing.",
										"TESTIFY_NS_NAT_PORTMAPED, // We are behind a NAT network with port forwarding, but no global IP.",
										"TESTIFY_NS_NAT_PORTMAPED_GLOBAL, // We are behind a NAT network with port forwarding and global IP.",
										"TESTIFY_NS_GLOBAL_DHCP_IP, // We have a global DHCP address.",
										"TESTIFY_NS_GLOBAL_DHCP_IP_UNSTABLE, // We have a global address but it is very unstable.",
										"TESTIFY_NS_STATIC_IP // We have a static IP address"};

char *testify_service_type_string[] = {"TESTIFY_SST_ADAPTER", // IP Address of an installed adapter
										"TESTIFY_SST_GATEWAY", // IP Address of a gateway
										"TESTIFY_SST_BROADCAST", // Broadcast Address.
										"TESTIFY_SST_MAPPED_ADDRESS", // Mapped address
										"TESTIFY_SST_DNS_GIVEN", // DNS Address configured on the syste.
										"TESTIFY_SST_DNS", // Alternative DNS addresses
										"TESTIFY_SST_DNS_VERIFY", // Addresses used to verify that DNS works.
										"TESTIFY_SST_HTTP", // Http servers that return the global ip of the computer
										"TESTIFY_SST_IRC", // DNS server to known IRC servers.
										"TESTIFY_SST_MASTER_SERVER", // Addresses to DNS addresses.
										"TESTIFY_SST_COUNT"};

char *testify_service_status_string[] = {"TESTIFY_SS_INACTIVE",
										"TESTIFY_SS_TRIGGERD",
										"TESTIFY_SS_WAITING",
										"TESTIFY_SS_SUCSESS",
										"TESTIFY_SS_SUCSESS_DOWNGRADED",
										"TESTIFY_SS_FAILED",
										"TESTIFY_SS_COUNT"};

char *testify_service_status_flags_string[] = {"TESTIFY_NS_CONNECTED", // Do we have any IP network connection at all.
												"TESTIFY_NS_IPV4", // Do we have a IPV4 Address
												"TESTIFY_NS_IPV6", // Do we have a IPV6 Address
												"TESTIFY_NS_DNS", // Do we have DNS access
												"TESTIFY_NS_INTERNET_ACCESS_CONFIRMED", // Do we have confirmed internet access
												"TESTIFY_NS_LAN_ACCESS", // Are we on local lan									
												"TESTIFY_NS_GLOBAL_IP_V4", // Does this machin have aglobal IP V4			
												"TESTIFY_NS_GLOBAL_IP_V6", // Does this machin have aglobal IP V6
												"TESTIFY_NS_GLOBAL_IP_CONFIRMED", // Is our global or NAT global IP Confirmed
												"TESTIFY_NS_BEHIND_NAT", // Are we behind a NAT
												"TESTIFY_NS_BEHIND_LOGIN", // Are we behind a NAT
												"TESTIFY_NS_PORT_FORWARDING_AVAILABLE", // Is any form of port forwarding available
												"TESTIFY_NS_PORTMAPED_PCP", // The nat we are behind supports PCP
												"TESTIFY_NS_PORTMAPED_PMP"}; // The nat we are behind supports PMP


void testify_status_debug_print()
{
	uint i;
	imagine_mutex_lock(testify_status.mutex);
	printf("Testify Debug status:\n");
	printf("---------------------\n");
	for(i = 0; i < testify_status.service_count; i++)
	{
		if(testify_status.services[i].status != TESTIFY_SS_INACTIVE)
		{
			printf("Name: %s\n", testify_status.services[i].name);
			printf("\tType %s\n", testify_service_type_string[testify_status.services[i].type]);
			if(testify_status.services[i].status != TESTIFY_SS_SUCSESS)
			{
				printf("\tStatus %s\n", testify_service_status_string[testify_status.services[i].status]);
			}else
			{
				if(testify_status.services[i].address.transport != TESTIFY_TRANSPORT_IPV4 ||
					testify_status.services[i].address.address.ipv4 != 0)
				{
					char buffer[TESTIFY_ADDRESS_MAX_STRING_SIZE];
					testify_address_print(&testify_status.services[i].address, buffer);
					printf("\tAddress %s\n", buffer);
				}
			}
		}
	}
	for(i = 0; i < TESTIFY_NS_COUNT; i++)
	{
		if((testify_status.status_bits >> i) & 1)
			printf("- %s TRUE\n", testify_service_status_flags_string[i]);
		else
			printf("- %s FALSE\n", testify_service_status_flags_string[i]);
	}
	imagine_mutex_unlock(testify_status.mutex);
}

uint64 testify_status_flags_get(void)
{
	uint64 flags;
	imagine_mutex_lock(testify_status.mutex);
	flags = testify_status.status_bits;
	imagine_mutex_unlock(testify_status.mutex);
	return flags;
}

void testify_status_flags_set(uint64 flags, uint64 mask)
{
	imagine_mutex_lock(testify_status.mutex);
	testify_status.status_bits &= ~mask;
	testify_status.status_bits |= flags;
	imagine_mutex_unlock(testify_status.mutex);
}
 

extern void testify_adapter_thread(void *user);
extern void testify_status_thread_func(void *user);
extern void testify_adapter_update();

void testify_status_initialize()
{
	if(!testify_status_init)
	{
		TestifyAddress address;
		testify_status_init = TRUE;
		testify_status.mutex = imagine_mutex_create();
		address.transport = TESTIFY_TRANSPORT_IPV4;
		address.address.ipv4 = 0;//testify_ipv4_address(224, 0, 0, 1);
	//	address.port = 5350;
		address.port = 0;
		testify_status.status_bits = 0;
		testify_status.handle = testify_create(&address, TESTIFY_PROTOCOL_DATAGRAM, TESTIFY_FLAG_BIND);
	/*	if(testify_status.handle == NULL)
		{
			address.address.ipv4 = testify_ipv4_address(127, 0, 0, 1);
			testify_status.handle = testify_create(&address, TESTIFY_PROTOCOL_DATAGRAM, TESTIFY_FLAG_BIND);
			if(testify_status.handle == NULL)
				testify_status.handle = testify_create(&address, TESTIFY_PROTOCOL_DATAGRAM, 0);
		}*/
	
		testify_adapter_update();
#ifdef _WIN32
		imagine_thread(testify_adapter_thread, NULL, "Testify adapter update");
#endif
		imagine_thread(testify_status_thread_func, NULL, "Testify Status");
	}
}

void testify_status_adapter_reset()
{
	uint i;
	testify_status_initialize();
	imagine_mutex_lock(testify_status.mutex);
	for(i = 0; i < testify_status.service_count; i++)
	{
		if(testify_status.services[i].type == TESTIFY_SST_ADAPTER ||
			testify_status.services[i].type == TESTIFY_SST_GATEWAY ||
			testify_status.services[i].type == TESTIFY_SST_BROADCAST ||
			testify_status.services[i].type == TESTIFY_SST_DNS_GIVEN)
			testify_status.services[i--] = testify_status.services[--testify_status.service_count];
	}
	imagine_mutex_unlock(testify_status.mutex);
}

void testify_status_service_set(TestifyServiceType type, char *name, TestifyAddress *address)
{
	TestifyStatusService *s;
	testify_status_initialize();
	imagine_mutex_lock(testify_status.mutex);
	if(testify_status.service_count == testify_status.service_allocated)
	{
		testify_status.service_allocated += 16;
		s = realloc(testify_status.services, (sizeof *testify_status.services) * testify_status.service_allocated);
		if(s == NULL)
		{
			imagine_mutex_unlock(testify_status.mutex);
			return;
		}
		testify_status.services = s;
	}
	s = &testify_status.services[testify_status.service_count++];

	if(type == TESTIFY_SST_GATEWAY)	
		s->status = TESTIFY_SS_WAITING; // NATPCP Protocol id
	else if(type == TESTIFY_SST_ADAPTER ||
		type == TESTIFY_SST_BROADCAST ||
		type == TESTIFY_SST_DNS_GIVEN)
		s->status = TESTIFY_SS_SUCSESS;
	else
		s->status = TESTIFY_SS_INACTIVE;
	//	s->status = TESTIFY_SS_TRIGGERD;
	s->type = type;
	s->name = f_text_copy_allocate(name);
	if(address != NULL)
	{
		s->address = *address;
	}else
	{
		s->address.transport = TESTIFY_TRANSPORT_IPV4;
		s->address.address.ipv4 = 0;
		s->address.port = 0;
	}
	s->last_updated = 0;
	s->first_detected = 0;
	s->handle = NULL;
	imagine_mutex_unlock(testify_status.mutex);
}


boolean	testify_service_is_localhost(TestifyAddress *address)
{
	TestifyAddress a;
	uint i;
	a = *address;
	imagine_mutex_lock(testify_status.mutex);
	for(i = 0; i < testify_status.service_count; i++)
	{
		if(testify_status.services[i].type == TESTIFY_SST_ADAPTER ||
			testify_status.services[i].type == TESTIFY_SST_HTTP)
		{
			a.port = testify_status.services[i].address.port;
			if(testify_address_compare(&testify_status.services[i].address, &a))
			{
				imagine_mutex_unlock(testify_status.mutex);
				return TRUE;
			}
		}
	}
	imagine_mutex_unlock(testify_status.mutex);
	return FALSE;
}

boolean testify_service_available(TestifyServiceType type, TestifyAddress *address)
{
	TestifyAddress a;
	uint i;
	a = *address;
	for(i = 0; i < testify_status.service_count; i++)
	{
		if(testify_status.services[i].type == type)
		{
			a.port = testify_status.services[i].address.port;
			if(testify_address_compare(&testify_status.services[i].address, &a))
				return TRUE;
		}
	}
	return FALSE;
}

boolean	testify_status_service_get(uint nr, TestifyServiceType type, char *name, uint buffer_size, TestifyAddress *address, TestifyServiceState *status)
{
	uint i, j;
	imagine_mutex_lock(testify_status.mutex);
	for(i = j = 0; i < testify_status.service_count; i++)
	{
		if(testify_status.services[i].type == type)
		{
			if(j++ == nr)
			{
				if(address !=  NULL)
					*address = testify_status.services[i].address;
				if(name != NULL && buffer_size != 0)
				{
					buffer_size--;
					for(j = 0; j < buffer_size && testify_status.services[i].name[j] != 0; j++)
						name[j] = testify_status.services[i].name[j];
					name[j] = 0;
				}
				if(status !=  NULL)
					*status = testify_status.services[i].status;
				imagine_mutex_unlock(testify_status.mutex);
				return TRUE;
			}
		}
	}
	imagine_mutex_unlock(testify_status.mutex);
	return FALSE;
}

void testify_status_initialize_default(boolean activate)
{
	char *irc_server[11] = {"chat.freenode.net", "irc.dal.net", "irc.gamesurge.net", "open.ircnet.net", "irc.quakenet.org", "irc.rizon.net", "irc.swiftirc.net", "eu.undernet.org", "us.undernet.org", "irc.webchat.org", "irc.2600.net"};
	char *ip_http_lookups[7] = {"myip.dnsomatic.com", "ifconfig.me", "ident.me", "icanhazip.com", "ipecho.net/plain", "checkip.amazonaws.com", "whatismyip.akamai.com"};
	char *verify_dns[7] = {"google.com", "facebook.com", "youtube.com", "aamazon.com", "wikipedia.org", "microsoft.com", "apple.com"};
	testify_status_initialize();
	uint i, start;
	start = testify_status.service_count;
	for(i = 0; i < 11; i++)
		testify_status_service_set(TESTIFY_SST_IRC, irc_server[i], NULL);
	for(i = 0; i < 7; i++)
		testify_status_service_set(TESTIFY_SST_HTTP, ip_http_lookups[i], NULL);
	for(i = 0; i < 7; i++)
		testify_status_service_set(TESTIFY_SST_DNS_VERIFY, verify_dns[i], NULL);
	if(activate)
		for(i = start; i < testify_status.service_count; i++)
			if(testify_status.services[i].type != TESTIFY_SST_IRC)
				testify_status.services[i].status = TESTIFY_SS_TRIGGERD;
}

uint testify_status_service_select(TestifyServiceType type, char *name, uint buffer_size, TestifyAddress *address, TestifyServiceState status)
{
	uint i, j;
	imagine_mutex_lock(testify_status.mutex);
	for(i = 0; i < testify_status.service_count; i++)
	{
		if(testify_status.services[i].type == type && testify_status.services[i].status == status)
		{
			if(address !=  NULL)
				*address = testify_status.services[i].address;
			if(name != NULL && buffer_size != 0)
			{
				buffer_size--;
				for(j = 0; j < buffer_size && testify_status.services[i].name[j] != 0; j++)
					name[j] = testify_status.services[i].name[j];
				name[j] = 0;
			}
			imagine_mutex_unlock(testify_status.mutex);
			return i;
		}
	}
	imagine_mutex_unlock(testify_status.mutex);
	return ~0;
}

void testify_status_service_status_set(TestifyServiceType type, char *name, TestifyServiceState status)
{
	uint i, j;
	imagine_mutex_lock(testify_status.mutex);
	for(i = 0; i < testify_status.service_count; i++)
	{
		if(testify_status.services[i].type == type)
		{
			for(j = 0; name[j] != 0 && name[j] == testify_status.services[i].name[j]; j++);
			if(name[j] == testify_status.services[i].name[j])
				testify_status.services[i].status = status;
		}
	}
	imagine_mutex_unlock(testify_status.mutex);
}

extern void testify_portmap_send(TestifyHandle *handle, TestifyAddress *gateway, TestifyProtocol protocol, uint version, uint timer, TestifyAddress *forward);



void testify_status_portmap_set(TestifyHandle *handle)
{
	void *mapped;
	uint i, count;
	testify_status_initialize();
	imagine_mutex_print_debug();
	imagine_mutex_lock(testify_status.mutex);
	mapped = testify_status.mapped_ports;
	handle->next_mapped_handle = testify_status.mapped_ports;
	testify_status.mapped_ports = handle;
	imagine_mutex_unlock(testify_status.mutex);
		
}

void testify_status_portmap_remove(TestifyHandle *handle)
{
	TestifyHandle **h;
	if(testify_status.mutex == NULL)
		return;
	imagine_mutex_lock(testify_status.mutex);
	for(h = &testify_status.mapped_ports; (*h) != NULL; h = &(*h)->next_mapped_handle)
	{
		if(*h == handle)
		{
			*h = handle->next_mapped_handle;
			break;
		}
	}
	imagine_mutex_unlock(testify_status.mutex);
}

boolean testify_address(TestifyHandle *handle, uint id, TestifyAddress *output)
{
	TestifyAddress a;
	uint i, j = 1;
	testify_status_initialize();
	imagine_mutex_lock(testify_status.mutex);
	if(handle->bound_to_specific_ip)
	{
		if(id == 0)
			*output = handle->address;
	}else
	{
		for(i = j = 0; i < testify_status.service_count; i++)
		{
			if(testify_status.services[i].type == TESTIFY_SST_ADAPTER)
			{
				if(j++ == id)
				{				
					*output = testify_status.services[i].address;
					output->port = handle->address.port;
					imagine_mutex_unlock(testify_status.mutex);
					return TRUE;
				}
			}
		}
	}

	for(i = 0; i < handle->mapping_count; i++)
	{
		if(handle->mappings[i].address.transport != TESTIFY_TRANSPORT_NULL)
		{
			if(j++ == id)
			{
				char buffer[TESTIFY_ADDRESS_MAX_STRING_SIZE];
				testify_address_print(&handle->mappings[i].address, buffer);
				printf("mapped address %s\n", buffer);
				*output = handle->mappings[i].address;	
				imagine_mutex_unlock(testify_status.mutex);
				return TRUE;
			}
		}
	}
	imagine_mutex_unlock(testify_status.mutex);
	return FALSE;
}

TestifyProtocol	testify_protocol(TestifyHandle *handle)
{
	return handle->protocol;
}

#define testify_ipv4_address(a, b, c, d) (a * (256 * 256 * 256) + b * (256 * 256) + c * 256 + d)


boolean testify_portmap_analyse_stable()
{
/*	uint i;
	for(i = 0; i < TESTIFY_PORTMAPPING_IP_COUNT; i++)
		if(testify_mapped_ports.global_addresses[i].address.transport == TESTIFY_TRANSPORT_COUNT)
			return TRUE;
	for(i = 0; i < TESTIFY_PORTMAPPING_IP_COUNT; i++)
	{
		if(testify_mapped_ports.global_addresses[i].last_updated < TESTIFY_PORTMAPPING_DNS_SLOW_TIMER &&
			testify_mapped_ports.global_addresses[i].first_detected > TESTIFY_PORTMAPPING_DNS_SLOW_TIMER * 8.0)
			return TRUE;
	}*/
	return FALSE;
}

void testify_portmap_analyse()
{
/*	uint i, j, adapter_count;
	if(testify_mapped_ports.nat_address.transport != TESTIFY_TRANSPORT_IPV4 || testify_mapped_ports.nat_address.address.ipv4 != 0)
	{
		for(i = 0; i < TESTIFY_PORTMAPPING_IP_COUNT && !testify_address_compare(&testify_mapped_ports.global_addresses[i].address, &testify_mapped_ports.nat_address); i++);
		if(i < TESTIFY_PORTMAPPING_IP_COUNT)
		{
			testify_mapped_ports.status = TESTIFY_NS_NAT_PORTMAPED_GLOBAL;
		}else
		{
			if(testify_portmap_analyse_stable())
				testify_mapped_ports.status = TESTIFY_NS_NAT_PORTMAPED;
			else
				testify_mapped_ports.status = TESTIFY_NS_NAT_PORTMAPED_UNSTABLE;
		}
	}else
	{
		for(i = 0; i < TESTIFY_PORTMAPPING_IP_COUNT && testify_mapped_ports.global_addresses[i].address.transport != TESTIFY_TRANSPORT_COUNT && testify_mapped_ports.global_addresses[i].last_updated < TESTIFY_PORTMAPPING_DNS_SLOW_TIMER * 4.0; i++);
		if(i == TESTIFY_PORTMAPPING_IP_COUNT)
		{
			testify_mapped_ports.status = TESTIFY_NS_UNKNOWN;
		}else
		{
			adapter_count = testify_network_adapter_count();
			for(i = 0; i < adapter_count; i++)
			{
				for(j = 0; j < TESTIFY_PORTMAPPING_IP_COUNT; j++)
				{
					if(testify_mapped_ports.global_addresses[i].address.transport != TESTIFY_TRANSPORT_IPV4 != 0 && testify_mapped_ports.global_addresses[i].address.address.ipv4 == testify_network_adapter_ipv4_local_address(j))
					{

						if(testify_portmap_analyse_stable())
							testify_mapped_ports.status = TESTIFY_NS_GLOBAL_DHCP_IP;
						else
							testify_mapped_ports.status = TESTIFY_NS_GLOBAL_DHCP_IP_UNSTABLE;
						return;
					//	TESTIFY_NS_STATIC_IP  FIX ME
					}
				}
			}

			if(testify_portmap_analyse_stable())
				testify_mapped_ports.status = TESTIFY_NS_NAT;
			else
				testify_mapped_ports.status = TESTIFY_NS_NAT_UNSTABLE;
		}
	}*/
}


extern boolean testify_http_whoami(char *host, TestifyAddress *output_address);
extern boolean testify_dns_send_buffer(TestifyHandle *handle, uint service);
extern boolean testify_dns_parse(uint8 *buffer, uint buffer_size, TestifyAddress *from);
extern void testify_adapter_update();

void testify_status_thread_func(void *user)
{
	static seconds = 0, fractions = 0;
	TestifyDiscovery *discover;
	TestifyAddress from, to, address_copy;
	uint i, j, k, size, time, adapter_count;
	uint8 buffer[1500];
	uint16 port, public_port;
	uint32 s, f, tmp;
	boolean sleep_timer = FALSE, test;
	double delta, addapter_timer = 100000.0;
	char *name_copy;
	imagine_current_time_get(&seconds, &fractions);
	while(TRUE)
	{
		imagine_current_time_get(&s, &f);
		delta = imagine_delta_time_compute(s, f, seconds, fractions);
		imagine_mutex_lock(testify_status.mutex);

		if(s != seconds)
			testify_portmap_update(&testify_status);
		seconds = s;
		fractions = f;
		sleep_timer = FALSE;

		for(i = 0; i < testify_status.service_count; i++)
		{
			testify_status.services[i].first_detected += delta;
			testify_status.services[i].last_updated += delta;
		}

		for(i = 0; i < testify_status.service_count; i++)
		{
			if(testify_status.services[i].status == TESTIFY_SS_TRIGGERD)
			{
				testify_status.services[i].status = TESTIFY_SS_WAITING;
				switch(testify_status.services[i].type)
				{
					case TESTIFY_SST_ADAPTER : // IP Address of an installed adapter
					break;
					case TESTIFY_SST_GATEWAY : // IP Address of a gateway
					break;
					case TESTIFY_SST_BROADCAST :
					break;
					case TESTIFY_SST_DNS_GIVEN : // DNS Address configured on the syste.
					break;
					case TESTIFY_SST_DNS : // Alternative DNS addresses
					break;
					case TESTIFY_SST_DNS_VERIFY : // Addresses used to verify that DNS works.
						address_copy = testify_status.services[i].address;
						name_copy = f_text_copy_allocate(testify_status.services[i].name);
						imagine_mutex_unlock(testify_status.mutex);
						test = testify_address_lookup(&address_copy, name_copy, TESTIFY_TRANSPORT_IPV4, 80);
						if(name_copy != NULL)
							free(name_copy);
						imagine_mutex_lock(testify_status.mutex);						
						if(i < testify_status.service_count && testify_address_compare(&testify_status.services[i].address, &address_copy))
						{
							if(test)
								testify_status.services[i].status = TESTIFY_SS_SUCSESS;
							else
								testify_status.services[i].status = TESTIFY_SS_FAILED;
						}
					break;
					case TESTIFY_SST_HTTP : // Http servers that return the global ip of the computer
						address_copy = testify_status.services[i].address;
						name_copy = f_text_copy_allocate(testify_status.services[i].name);
						imagine_mutex_unlock(testify_status.mutex);
						test = testify_http_whoami(name_copy, &address_copy);
						if(name_copy != NULL)
							free(name_copy);
						imagine_mutex_lock(testify_status.mutex);						
						if(i < testify_status.service_count && testify_address_compare(&testify_status.services[i].address, &address_copy))
						{
							if(test)
							{
								testify_status.services[i].status = TESTIFY_SS_SUCSESS;

							}else
								testify_status.services[i].status = TESTIFY_SS_FAILED;
						}
					break;
					case TESTIFY_SST_IRC : // DNS server to known IRC servers.
					break;
					case TESTIFY_SST_MASTER_SERVER : // Addresses to DNS addresses.
					break;
				}

				sleep_timer = TRUE;
				break;
			}
		}
		
		imagine_mutex_unlock(testify_status.mutex);
		size = testify_datagram_receive(testify_status.handle, &from, buffer, 1500);
		if(size == 0)
		{	
			if(sleep_timer)
				imagine_sleepd(0.01);
			else
				imagine_sleepd(0.1);
		}
#ifndef _WIN32
		addapter_timer += delta;
		if(addapter_timer > 5.0)
		{
			testify_adapter_update();
			addapter_timer = 0;
		}
#endif
		if(size != 0)
		{
			if(from.port == 5351 && size >= 4)
			{
				imagine_mutex_lock(testify_status.mutex);
				testify_portmap_recive(&testify_status, &from, buffer, size);
				imagine_mutex_unlock(testify_status.mutex);
			}
			if(from.port == 53 && size >= 4)
			{
		/*		i = 0;
				if(testify_portmap_dns_parse(buffer, size, &to))
				{
					if(to.transport == TESTIFY_TRANSPORT_IPV4)
						testify_portmap_global_address_update(to);
				}*/
			}
		}
		i  = 0;
		for(i = 0; TRUE; i++)
		{
			imagine_mutex_lock(testify_status.mutex);
			j = 0;
			for(discover = testify_status.discovery; j < i && discover != NULL; discover = discover->next)
				j++;
			if(discover != NULL)
			{
				imagine_mutex_lock(discover->mutex);
				imagine_mutex_unlock(testify_status.mutex);
				testify_discover_update(discover, delta);
				imagine_mutex_unlock(discover->mutex);
			}else
			{
				imagine_mutex_unlock(testify_status.mutex);
				break;
			}
		}
	}
}
