#ifdef TEST
#pragma makemake name testify_test
#include <stdio.h>
#include "imagine.h"
#include "testify.h"

void testify_test_address_print(TestifyHandle *handle)
{	
	char *type_name[] = {"LOOPBACK", "PRIVATE NETWORK", "DOCUMENTATION", "BROADCAST", "MULTICAST", "GLOBAL"};
	char address_string[TESTIFY_ADDRESS_MAX_STRING_SIZE];
	uint i, type;
	TestifyAddress address;
	if(handle == NULL)
		return;		
	printf("Address:\n");
	for(i = 0; testify_address(handle, i, &address); i++)
	{
		testify_address_print(&address, address_string);
		type = testify_address_type(&address);
		printf("\t%s (%s)\n", address_string, type_name[type]);
	}
}

void testify_test_discover_list_print(TestifyDiscovery *discover)
{
	char address_string[TESTIFY_ADDRESS_MAX_STRING_SIZE];
	TestifyDiscoverPeer *peers;
	uint i, peer_count;
	peers = testify_discover_service_list_get(discover, &peer_count);
	for(i = 0; i < peer_count; i++)
	{
		testify_address_print(&peers[i].address, address_string);
		printf("\tAddress: %s %s -", address_string, peers[i].message);
		if(peers[i].method & TESTIFY_DMF_BROADCAST_LOCAL)
			printf(" local");
		if(peers[i].method & TESTIFY_DMF_BROADCAST_LAN)
			printf(" lan");
		if(peers[i].method & TESTIFY_DMF_BROADCAST_WAN)
			printf(" wan");
		if(peers[i].method & TESTIFY_DMF_WAN_IRC)
			printf(" irc");
		printf("\n", address_string, peers[i].message);

	}
	if(peer_count == 0)
		printf("\tNo Peers found\n", address_string, peers[i].message);
	if(peers != NULL)
		free(peers);
}

extern void testify_print_network_status_debug();
extern TestifyHandle *testify_irc_parser(TestifyHandle *handle, void (*command_funct)(uint command, char *sender, char *channel, char *mesassage, void *user), void *user);
extern testify_discover_irc_print();

void testify_command_func(uint command, char *sender, char *channel, char *mesassage)
{
	printf("Command %u", command);
	if(sender != NULL)
		printf(" sender: %s", sender);
	if(channel != NULL)
		printf(" channel: %s", channel);
	if(mesassage != NULL)
		printf(" message: %s", mesassage);
	printf("\n");
}

extern int main_test(char *p);

int main(int argc, char **argv)
{
	TestifyHandle *udp_ipv4_handle;
	TestifyHandle *udp_ipv6_handle;
	TestifyHandle *tcp_ipv4_handle;
	TestifyHandle *tcp_ipv6_handle;
	TestifyHandle *bluetooth_handle;
	TestifyHandle *irc_handle = NULL;
	TestifyAddress address;
	TestifyDiscovery *discover_broadcast = NULL;
	TestifyDiscovery *discover_global = NULL;
	TestifyDiscovery *discover_bluetooth = NULL;
	char peer_name[32] = {0};
	uint i, j, k;
//	main_test("C:\\Sorted\\");
//	main_test("C:\\Sorted\\Riot\\");

	
	testify_network_adapter_count();
	testify_status_initialize_default(TRUE);
	printf("TESTIFY TEST\n");
	printf("------------\n");
	printf("This application tests Testifyes features and diagnoses the network\n");
	printf("Please enter your peer name:\n");
	scanf("%31s", peer_name);
	printf("Name: %s\n\n", peer_name);

	if(testify_address_lookup(&address, "any", TESTIFY_TRANSPORT_IPV4, 0))
	{
		udp_ipv4_handle = testify_create(&address, TESTIFY_PROTOCOL_DATAGRAM, TESTIFY_FLAG_BIND | TESTIFY_FLAG_PORT_MAPPED);
		if(udp_ipv4_handle == NULL)
			printf("Testify Error: Failed to bind port for UDP IPV4 handle.\n");
	}else
	{
		printf("Testify Error: Failed to aquire address for UDP IPV4 handle.\n");
		udp_ipv4_handle = NULL;
	}
	
	if(testify_address_lookup(&address, "any", TESTIFY_TRANSPORT_IPV6, 0))
	{
		udp_ipv6_handle = testify_create(&address, TESTIFY_PROTOCOL_DATAGRAM, TESTIFY_FLAG_BIND | TESTIFY_FLAG_PORT_MAPPED);
		if(udp_ipv6_handle == NULL)
			printf("Testify Error: Failed to bind port for UDP IPV6 handle.\n");
	}else
	{
		printf("Testify Error: Failed to aquire address for UDP IPV6 handle.\n");
		udp_ipv6_handle = NULL;
	}

	if(testify_address_lookup(&address, "any", TESTIFY_TRANSPORT_IPV4, 0))
	{
		tcp_ipv4_handle = testify_create(&address, TESTIFY_PROTOCOL_STREAM, TESTIFY_FLAG_BIND | TESTIFY_FLAG_PORT_MAPPED);
		if(tcp_ipv4_handle == NULL)
			printf("Testify Error: Failed to bind port for TCP IPV4 handle.\n");
	}else
	{
		printf("Testify Error: Failed to aquire address for TCP IPV4 handle.\n");
		tcp_ipv4_handle = NULL;
	}
	
	if(testify_address_lookup(&address, "any", TESTIFY_TRANSPORT_IPV6, 0))
	{
		tcp_ipv6_handle = testify_create(&address, TESTIFY_PROTOCOL_STREAM, TESTIFY_FLAG_BIND | TESTIFY_FLAG_PORT_MAPPED);
		if(tcp_ipv6_handle == NULL)
			printf("Testify Error: Failed to bind port for TCP IPV6 handle.\n");
	}else
	{
		printf("Testify Error: Failed to aquire address for TCP IPV4 handle.\n");
		tcp_ipv6_handle = NULL;
	}

	if(testify_address_lookup(&address, "localhost", TESTIFY_TRANSPORT_BLUETOOTH, 4))
	{
		bluetooth_handle = testify_create(&address, TESTIFY_PROTOCOL_STREAM, TESTIFY_FLAG_BIND | TESTIFY_FLAG_PORT_MAPPED);
		if(bluetooth_handle == NULL)
			printf("Testify Error: Failed to bind port for Blutooth handle.\n");
	}else
	{
		printf("Testify Error: Failed to aquire address for Blutooth handle.\n");
		bluetooth_handle = NULL;
	}
	testify_status_debug_print();

/*	while(TRUE)
	{
		imagine_sleepd(2.0);
		testify_status_debug_print();
		printf("UDP ipv4:");
		testify_test_address_print(udp_ipv4_handle);
		printf("UDP ipv6:");
		testify_test_address_print(udp_ipv6_handle);
		printf("TCP ipv4:");
		testify_test_address_print(tcp_ipv4_handle);
		printf("TCP ipv6:");
		testify_test_address_print(tcp_ipv6_handle);
		printf("Bluetooth:");
		testify_test_address_print(bluetooth_handle);
		printf("\nStarting services:\n");
	}*/

//	discover_broadcast = testify_discover_create("Testify LAN", udp_ipv4_handle, peer_name, TESTIFY_DMF_BROADCAST_LOCAL | TESTIFY_DMF_BROADCAST_LAN);
	discover_global = testify_discover_create("Testify Wan", tcp_ipv4_handle, peer_name, /*TESTIFY_DMF_BROADCAST_LOCAL | TESTIFY_DMF_BROADCAST_LAN |*/ TESTIFY_DMF_WAN_IRC);
//	discover_bluetooth = testify_discover_create("Seduce", 18686, peer_name, TESTIFY_DMF_BUETOOTH);
//	irc_handle = testify_irc_create("irc.efnet.org", "testify_test", "quelsolaar");
	for(i = 0; i < 12000000; i++)
	{
		imagine_sleepd(1.0);
	/*	if(discover_broadcast != NULL && testify_discover_updated(discover_broadcast))
		{
			printf("Local update:\n");
			testify_test_discover_list_print(discover_broadcast);
		}*/
		if(discover_global != NULL && testify_discover_updated(discover_global))
		{
			printf("Global update:\n");
			testify_test_discover_list_print(discover_global);
		}
	/*	if(testify_discover_updated(discover_bluetooth))
		{
			printf("Bluetooth update:\n");
			testify_test_discover_list_print(discover_bluetooth);
		}
		if(i % 5 == 4)
		{
			testify_test_address_print(udp_ipv4_handle);
			testify_status_debug_print();
		}*/
		if(irc_handle != NULL)
			irc_handle = testify_irc_parser(irc_handle, testify_command_func, NULL);
		/*{
			char buffer[1501];
			size_t size;
			size = testify_stream_recive(irc_handle, buffer, 1500);
			buffer[size] = 0;
			if(size != 0)
				printf("%s", buffer);
		}*/
		if(irc_handle != NULL && i % 5 == 0)
		{
			for(j = 0; j < testify_irc_channel_count_get(irc_handle); j++)
			{
				printf("channel: %s\n", testify_irc_channel_name_get(irc_handle, j));
				for(k = 0; k < testify_irc_channel_user_count_get(irc_handle, j); k++)
				{
					printf("\tuser: %s\n", testify_irc_channel_user_name_get(irc_handle, j, k));
				}
				printf("Messages: \n");
				for(k = testify_irc_channel_message_count_get(irc_handle, j); k > 0; k--)
					printf("\t%s: %s\n", testify_irc_channel_message_user_name_get(irc_handle, j, k - 1), testify_irc_channel_message_get(irc_handle, j, k - 1));

			}
		}
	}

	testify_discover_destroy(discover_broadcast);
	while(TRUE)
		imagine_sleepd(1.0);
	return TRUE;
}
#endif